import numpy as np

from nengo._vendor.npconv2d import conv2d
from nengo.builder.builder import Builder
from nengo.builder.operator import DotInc, ElementwiseInc, Operator, Reset, SparseDotInc
from nengo.builder.signal import Signal
from nengo.exceptions import BuildError
from nengo.rc import rc
from nengo.transforms import (
    Convolution,
    ConvolutionTranspose,
    Dense,
    NoTransform,
    Sparse,
)


def multiply(x, y):
    """Matrix-matrix multiply, interpreting vectors as diagonal matrices."""
    if x.ndim <= 2 and y.ndim < 2:
        return x * y
    elif x.ndim < 2 and y.ndim == 2:
        return x.reshape(-1, 1) * y
    elif x.ndim == 2 and y.ndim == 2:
        return np.dot(x, y)
    else:
        raise BuildError(f"Tensors not supported (x.ndim={x.ndim}, y.ndim={y.ndim})")


@Builder.register(Dense)
def build_dense(model, transform, sig_in, decoders=None, encoders=None, rng=np.random):
    """Build a `.Dense` transform object."""

    weights = transform.sample(rng=rng).astype(rc.float_dtype)

    if decoders is not None:
        weights = multiply(weights, decoders.astype(rc.float_dtype))
    if encoders is not None:
        weights = multiply(encoders.astype(rc.float_dtype).T, weights)

    # Add operator for applying weights
    weight_sig = Signal(weights, readonly=True, name=f"{transform}.weights")
    weighted = Signal(
        shape=transform.size_out if encoders is None else weights.shape[0],
        name=f"{transform}.weighted",
    )
    model.add_op(Reset(weighted))

    op = ElementwiseInc if weights.ndim < 2 else DotInc
    model.add_op(op(weight_sig, sig_in, weighted, tag=f"{transform}.apply_weights"))

    return weighted, weight_sig


@Builder.register(Sparse)
def build_sparse(model, transform, sig_in, decoders=None, encoders=None, rng=np.random):
    """Build a `.Sparse` transform object."""

    if decoders is not None:
        raise BuildError(
            "Applying a sparse transform to a decoded connection is not supported"
        )

    # Shouldn't be possible for encoders to be non-None, since that only
    # occurs for a connection solver with weights=True, and those can only
    # be applied to decoded connections (which are disallowed above)
    assert encoders is None

    # Add output signal
    weighted = Signal(shape=transform.size_out, name=f"{transform}.weighted")
    model.add_op(Reset(weighted))

    weights = transform.sample(rng=rng)
    assert weights.ndim == 2

    # Add operator for applying weights
    weight_sig = Signal(weights, name=f"{transform}.weights", readonly=True)
    model.add_op(
        SparseDotInc(weight_sig, sig_in, weighted, tag=f"{transform}.apply_weights")
    )

    return weighted, weight_sig


@Builder.register(Convolution)
@Builder.register(ConvolutionTranspose)
def build_convolution(
    model, transform, sig_in, decoders=None, encoders=None, rng=np.random
):
    """Build a `.Convolution` transform object."""

    if decoders is not None:
        raise BuildError(
            "Applying a convolution transform to a decoded "
            "connection is not supported"
        )

    # Shouldn't be possible for encoders to be non-None, since that only
    # occurs for a connection solver with weights=True, and those can only
    # be applied to decoded connections (which are disallowed above)
    assert encoders is None

    weights = transform.sample(rng=rng)
    weight_sig = Signal(weights, readonly=True, name=f"{transform}.weights")
    weighted = Signal(shape=transform.size_out, name=f"{transform}.weighted")
    model.add_op(Reset(weighted))

    op_class = (
        ConvTransposeInc if isinstance(transform, ConvolutionTranspose) else ConvInc
    )
    model.add_op(
        op_class(
            weight_sig, sig_in, weighted, transform, tag=f"{transform}.apply_weights"
        )
    )

    return weighted, weight_sig


class GeneralConvInc(Operator):
    """Apply convolutional weights to input signal.

    .. versionadded:: 3.2.0

    Parameters
    ----------
    W : Signal
        The convolutional weights (a.k.a. the kernel).
    X : Signal
        The input signal.
    Y : Signal
        Output signal to be incremented.
    conv : `~nengo.Convolution` or `~nengo.ConvolutionTranspose`
        The Convolution or ConvolutionTranspose transform being applied.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Attributes
    ----------
    W : Signal
        The convolutional weights.
    X : Signal
        The input signal.
    Y : Signal
        Output signal to be incremented.
    conv : `~nengo.Convolution` or `~nengo.ConvolutionTranspose`
        The Convolution or ConvolutionTranspose transform being applied.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Notes
    -----
    1. sets ``[]``
    2. incs ``[Y]``
    3. reads ``[W, X]``
    4. updates ``[]``
    """

    def __init__(self, W, X, Y, conv, tag=None):
        super().__init__(tag=tag)

        assert isinstance(conv, (Convolution, ConvolutionTranspose))

        self.conv = conv

        self.sets = []
        self.incs = [Y]
        self.reads = [W, X]
        self.updates = []

    @property
    def is_transpose(self):
        return isinstance(self.conv, ConvolutionTranspose)

    @property
    def W(self):
        return self.reads[0]

    @property
    def X(self):
        return self.reads[1]

    @property
    def Y(self):
        return self.incs[0]

    @property
    def _descstr(self):
        name = "convtranspose2d" if self.is_transpose else "conv2d"
        return f"{name}({self.W}, {self.X}) -> {self.Y}"

    def make_step(self, signals, dt, rng):
        if self.conv.dimensions > 2:
            # note: we raise the error here, rather than earlier, because
            # other backends might support different convolutions
            raise NotImplementedError("Convolution > 2D not supported")

        W = signals[self.W]
        X = signals[self.X]
        Y = signals[self.Y]
        pad = self.conv.padding.upper()
        stride = self.conv.strides
        output_spatial_shape = (
            self.conv.output_shape.spatial_shape if self.is_transpose else None
        )

        X = X.reshape(self.conv.input_shape.shape)
        Y = Y.reshape(self.conv.output_shape.shape)

        if not self.conv.channels_last:
            X = np.moveaxis(X, 0, -1)
            Y = np.moveaxis(Y, 0, -1)

        if self.conv.dimensions == 1:
            # add extra dimension to make it a 2D convolution
            X = X[None, :, :]
            W = W[None, :, :, :]
            Y = Y[None, :, :]
            stride = (1,) + stride
            if output_spatial_shape is not None:
                output_spatial_shape = (1,) + output_spatial_shape

        # add empty batch dimension
        X = X[None, ...]

        if self.is_transpose:

            def step_conv_transpose():
                Y[...] += conv2d.conv2d_gradx(
                    W, X, xsize=output_spatial_shape, pad=pad, stride=stride
                )[0]

            return step_conv_transpose

        else:

            def step_conv():
                Y[...] += conv2d.conv2d(X, W, pad=pad, stride=stride)[0]

            return step_conv


class ConvInc(GeneralConvInc):
    """Apply convolutional weights to input signal.

    .. versionadded:: 3.0.0

    Parameters
    ----------
    W : Signal
        The convolutional weights (a.k.a. the kernel).
    X : Signal
        The input signal.
    Y : Signal
        Output signal to be incremented.
    conv : `~nengo.Convolution`
        The Convolution transform being applied.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Attributes
    ----------
    W : Signal
        The convolutional weights.
    X : Signal
        The input signal.
    Y : Signal
        Output signal to be incremented.
    conv : `~nengo.Convolution`
        The Convolution transform being applied.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Notes
    -----
    1. sets ``[]``
    2. incs ``[Y]``
    3. reads ``[W, X]``
    4. updates ``[]``
    """

    def __init__(self, W, X, Y, conv, tag=None):
        super().__init__(W, X, Y, conv, tag=tag)
        assert not self.is_transpose


class ConvTransposeInc(GeneralConvInc):
    """Apply transposed convolutional weights to input signal.

    .. versionadded:: 3.2.0

    Parameters
    ----------
    W : Signal
        The convolutional weights (a.k.a. the kernel).
    X : Signal
        The input signal.
    Y : Signal
        Output signal to be incremented.
    conv : `~nengo.ConvolutionTranspose`
        The ConvolutionTranspose transform being applied.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Attributes
    ----------
    W : Signal
        The convolutional weights.
    X : Signal
        The input signal.
    Y : Signal
        Output signal to be incremented.
    conv : `~nengo.ConvolutionTranspose`
        The ConvolutionTranspose transform being applied.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Notes
    -----
    1. sets ``[]``
    2. incs ``[Y]``
    3. reads ``[W, X]``
    4. updates ``[]``
    """

    def __init__(self, W, X, Y, conv, tag=None):
        super().__init__(W, X, Y, conv, tag=tag)
        assert self.is_transpose


@Builder.register(NoTransform)
def build_no_transform(
    model, transform, sig_in, decoders=None, encoders=None, rng=np.random
):
    """Build a `.NoTransform` transform object."""

    if decoders is not None or encoders is not None:
        return build_dense(
            model,
            Dense(shape=(transform.size_out, transform.size_in), init=1.0),
            sig_in,
            decoders=decoders,
            encoders=encoders,
            rng=rng,
        )

    return sig_in, None



"""Helper functions for backends generating their own Builder system."""

from collections import defaultdict

import numpy as np

from nengo.exceptions import MovedError, Unconvertible, ValidationError


def full_transform(  # noqa: C901
    conn, slice_pre=True, slice_post=True, allow_scalars=True
):
    """Compute the full transform matrix for a Dense connection transform.

    Parameters
    ----------
    conn : Connection
        The connection for which to compute the full transform.
    slice_pre : boolean, optional (True)
        Whether to compute the pre slice as part of the transform.
    slice_post : boolean, optional (True)
        Whether to compute the post slice as part of the transform.
    allow_scalars : boolean, optional (True)
        If true (default), will not make scalars into full transforms when
        not using slicing, since these work fine in the reference builder.
        If false, these scalars will be turned into scaled identity matrices.
    """
    # imported here to avoid circular imports
    # pylint: disable=import-outside-toplevel
    from nengo.transforms import Dense, NoTransform

    if isinstance(conn.transform, NoTransform):
        transform = np.array(1.0)
    elif not isinstance(conn.transform, Dense):
        raise ValidationError(
            "full_transform can only be applied to Dense transforms",
            attr="transform",
            obj=conn,
        )
    else:
        transform = conn.transform.init

    pre_slice = conn.pre_slice if slice_pre and conn.function is None else slice(None)
    post_slice = conn.post_slice if slice_post else slice(None)

    eq_none_slice = lambda s: isinstance(s, slice) and s == slice(None)
    if eq_none_slice(pre_slice) and eq_none_slice(post_slice):
        if transform.ndim == 2:
            # transform is already full, so return a copy
            return np.array(transform)
        elif transform.size == 1 and allow_scalars:
            if transform.ndim == 1:
                return np.array(transform[0])
            else:
                return np.array(transform)

    # Create the new transform matching the pre/post dimensions
    func_size = conn.function_info.size
    size_in = (
        (conn.pre_obj.size_out if func_size is None else func_size)
        if slice_pre
        else conn.size_mid
    )
    size_out = conn.post_obj.size_in if slice_post else conn.size_out
    new_transform = np.zeros((size_out, size_in))

    # Dense transforms should not be able to have > 2 axes, but just in case
    assert transform.ndim <= 2, "connection transform must have <= 2 axes"

    if transform.ndim == 2:
        repeated_inds = lambda x: (
            not isinstance(x, slice) and np.unique(x).size != len(x)
        )
        if repeated_inds(pre_slice):
            raise NotImplementedError("Input object selection has repeated indices")
        if repeated_inds(post_slice):
            raise NotImplementedError("Output object selection has repeated indices")

        rows_transform = np.array(new_transform[post_slice])
        rows_transform[:, pre_slice] = transform
        new_transform[post_slice] = rows_transform
        # Note: the above is a little obscure, but we do it so that lists of
        #  indices can specify selections of rows and columns, rather than
        #  just individual items
    else:
        new_transform[
            np.arange(size_out)[post_slice], np.arange(size_in)[pre_slice]
        ] = transform
    return new_transform


def default_n_eval_points(n_neurons, dimensions):
    """A heuristic to determine an appropriate number of evaluation points.

    This is used by builders to generate a sufficiently large sample
    from a vector space in order to solve for accurate decoders.

    Parameters
    ----------
    n_neurons : int
        The number of neurons in the ensemble that will be sampled.
        For a connection, this would be the number of neurons in the
        ``pre`` ensemble.
    dimensions : int
        The number of dimensions in the ensemble that will be sampled.
        For a connection, this would be the number of dimensions in the
        ``pre`` ensemble.
    """
    from nengo.utils.numpy import clip  # pylint: disable=import-outside-toplevel

    return max(clip(500 * dimensions, 750, 2500), 2 * n_neurons)


def objs_and_connections(network):
    """Given a Network, returns all (ensembles + nodes, connections)."""
    return network.all_ensembles + network.all_nodes, network.all_connections


def generate_graphviz(*args, **kwargs):
    """Moved to nengo_extras.graphviz."""
    raise MovedError(location="nengo_extras.graphviz")


def _create_replacement_connection(c_in, c_out):
    """Generate a new Connection to replace two through a passthrough Node."""
    # imported here to avoid circular imports
    # pylint: disable=import-outside-toplevel
    from nengo.connection import Connection

    assert c_in.post_obj is c_out.pre_obj
    assert c_in.post_obj.output is None

    # determine the filter for the new Connection
    if c_in.synapse is None:
        synapse = c_out.synapse
    elif c_out.synapse is None:
        synapse = c_in.synapse
    else:
        raise Unconvertible("Cannot merge two filters")
        # Note: the algorithm below is in the right ballpark,
        #  but isn't exactly the same as two low-pass filters
        # filter = c_out.filter + c_in.filter

    function = c_in.function
    if c_out.function is not None:
        raise Unconvertible("Cannot remove a connection with a function")

    # compute the combined transform
    transform = np.dot(full_transform(c_out), full_transform(c_in))

    # check if the transform is 0 (this happens a lot
    #  with things like identity transforms)
    if np.all(transform == 0):
        return None

    c = Connection(
        c_in.pre_obj,
        c_out.post_obj,
        synapse=synapse,
        transform=transform,
        function=function,
        add_to_container=False,
    )
    return c


def remove_passthrough_nodes(  # noqa: C901
    objs, connections, create_connection_fn=None
):
    """Returns a version of the model without passthrough Nodes.

    For some backends (such as SpiNNaker), it is useful to remove Nodes that
    have 'None' as their output.  These nodes simply sum their inputs and
    use that as their output. These nodes are defined purely for organizational
    purposes and should not affect the behaviour of the model.  For example,
    the 'input' and 'output' Nodes in an EnsembleArray, which are just meant to
    aggregate data.

    Note that removing passthrough nodes can simplify a model and may be useful
    for other backends as well.  For example, an EnsembleArray connected to
    another EnsembleArray with an identity matrix as the transform
    should collapse down to D Connections between the corresponding Ensembles
    inside the EnsembleArrays.

    Parameters
    ----------
    objs : list of Nodes and Ensembles
        All the objects in the model
    connections : list of Connections
        All the Connections in the model

    Returns the objs and connections of the resulting model.  The passthrough
    Nodes will be removed, and the Connections that interact with those Nodes
    will be replaced with equivalent Connections that don't interact with those
    Nodes.
    """
    # imported here to avoid circular imports
    # pylint: disable=import-outside-toplevel
    from nengo.node import Node

    if create_connection_fn is None:
        create_connection_fn = _create_replacement_connection

    inputs, outputs = find_all_io(connections)
    result_conn = list(connections)
    result_objs = list(objs)

    # look for passthrough Nodes to remove
    for obj in objs:
        if isinstance(obj, Node) and obj.output is None:
            result_objs.remove(obj)

            # get rid of the connections to and from this Node
            for c in inputs[obj]:
                result_conn.remove(c)
                outputs[c.pre_obj].remove(c)
            for c in outputs[obj]:
                result_conn.remove(c)
                inputs[c.post_obj].remove(c)

            # replace those connections with equivalent ones
            for c_in in inputs[obj]:
                if c_in.pre_obj is obj:
                    raise Unconvertible(
                        "Cannot remove a Node with a feedback connection"
                    )

                for c_out in outputs[obj]:
                    c = create_connection_fn(c_in, c_out)
                    if c is not None:
                        result_conn.append(c)
                        # put this in the list, since it might be used
                        # another time through the loop
                        outputs[c.pre_obj].append(c)
                        inputs[c.post_obj].append(c)

    return result_objs, result_conn


def find_all_io(connections):
    """Build up a list of all inputs and outputs for each object."""
    inputs = defaultdict(list)
    outputs = defaultdict(list)
    for c in connections:
        inputs[c.post_obj].append(c)
        outputs[c.pre_obj].append(c)
    return inputs, outputs



import warnings

import numpy as np

import nengo.utils.numpy as npext
from nengo.exceptions import ConvergenceError, ValidationError
from nengo.params import (
    BoolParam,
    EnumParam,
    FrozenObject,
    IntParam,
    NdarrayParam,
    NumberParam,
    Parameter,
    Unconfigurable,
)
from nengo.utils.numpy import is_integer
from nengo.utils.paths import data_dir


class Distribution(FrozenObject):
    """A base class for probability distributions.

    The only thing that a probabilities distribution need to define is a
    `.Distribution.sample` method. This base class ensures that all
    distributions accept the same arguments for the sample function.
    """

    def _sample_shape(self, n, d=None):
        """Returns output shape for sample method."""
        return (n,) if d is None else (n, d)

    def sample(self, n, d=None, rng=np.random):
        """Samples the distribution.

        Parameters
        ----------
        n : int
            Number samples to take.
        d : int or None, optional
            The number of dimensions to return. If this is an int, the return
            value will be of shape ``(n, d)``. If None, the return
            value will be of shape ``(n,)``.
        rng : `numpy.random.RandomState`, optional
            Random number generator state.

        Returns
        -------
        samples : (n,) or (n, d) array_like
            Samples as a 1d or 2d array depending on ``d``. The second
            dimension enumerates the dimensions of the process.
        """
        raise NotImplementedError("Distributions should implement sample.")


class DistributionParam(Parameter):
    """A Distribution."""

    equatable = True

    def coerce(self, instance, dist):  # pylint: disable=arguments-renamed
        self.check_type(instance, dist, Distribution)
        return super().coerce(instance, dist)


class DistOrArrayParam(NdarrayParam):
    """Can be a Distribution or samples from a distribution."""

    def __init__(
        self,
        name,
        default=Unconfigurable,
        sample_shape=None,
        sample_dtype=np.float64,
        optional=False,
        readonly=None,
    ):
        super().__init__(
            name=name,
            default=default,
            shape=sample_shape,
            dtype=sample_dtype,
            optional=optional,
            readonly=readonly,
        )

    def coerce(self, instance, distorarray):  # pylint: disable=arguments-renamed
        if isinstance(distorarray, Distribution):
            return Parameter.coerce(self, instance, distorarray)
        return super().coerce(instance, distorarray)


def get_samples(dist_or_samples, n, d=None, rng=np.random):
    """Convenience function to sample a distribution or return samples.

    Use this function in situations where you accept an argument that could
    be a distribution, or could be an ``array_like`` of samples.

    Examples
    --------

    .. testcode::

       from nengo.dists import get_samples

       rng = np.random.RandomState(seed=0)

       def mean(values, n=100):
           samples = get_samples(values, n=n, rng=rng)
           print(f"{np.mean(samples):.4f}")

       mean([1, 2, 3, 4])
       mean(nengo.dists.Gaussian(0, 1))

    .. testoutput::

       2.5000
       0.0598

    Parameters
    ----------
    dist_or_samples : `.Distribution` or (n, d) array_like
        Source of the samples to be returned.
    n : int
        Number samples to take.
    d : int or None, optional
        The number of dimensions to return.
    rng : RandomState, optional
        Random number generator.

    Returns
    -------
    samples : (n, d) array_like

    """
    if isinstance(dist_or_samples, Distribution):
        return dist_or_samples.sample(n, d=d, rng=rng)
    return np.array(dist_or_samples)


class PDF(Distribution):
    """An arbitrary distribution from a PDF.

    Parameters
    ----------
    x : vector_like (n,)
        Values of the points to sample from (interpolated).
    p : vector_like (n,)
        Probabilities of the ``x`` points.
    """

    x = NdarrayParam("x", shape="*")
    p = NdarrayParam("p", shape="*")

    def __init__(self, x, p):
        super().__init__()

        psum = np.sum(p)
        if np.abs(psum - 1) > 1e-8:
            raise ValidationError(
                f"PDF must sum to one (sums to {psum:f})", attr="p", obj=self
            )

        self.x = x
        self.p = p
        if len(self.x) != len(self.p):
            raise ValidationError(
                "`x` and `p` must be the same length", attr="p", obj=self
            )

        # make cumsum = [0] + cumsum, cdf = 0.5 * (cumsum[:-1] + cumsum[1:])
        cumsum = np.cumsum(p)
        cumsum *= 0.5
        cumsum[1:] = cumsum[:-1] + cumsum[1:]
        self.cdf = cumsum

    def sample(self, n, d=None, rng=np.random):
        shape = self._sample_shape(n, d)
        return np.interp(rng.uniform(size=shape), self.cdf, self.x)


class Uniform(Distribution):
    """A uniform distribution.

    It's equally likely to get any scalar between ``low`` and ``high``.

    Note that the order of ``low`` and ``high`` doesn't matter;
    if ``low < high`` this will still work, and ``low`` will still
    be a closed interval while ``high`` is open.

    Parameters
    ----------
    low : Number
        The closed lower bound of the uniform distribution; samples >= low
    high : Number
        The open upper bound of the uniform distribution; samples < high
    integer : boolean, optional
        If true, sample from a uniform distribution of integers. In this case,
        low and high should be integers.
    """

    low = NumberParam("low")
    high = NumberParam("high")
    integer = BoolParam("integer")

    def __init__(self, low, high, integer=False):
        super().__init__()
        self.low = low
        self.high = high
        self.integer = integer

    def sample(self, n, d=None, rng=np.random):
        shape = self._sample_shape(n, d)
        if self.integer:
            return rng.randint(low=self.low, high=self.high, size=shape)
        else:
            return rng.uniform(low=self.low, high=self.high, size=shape)


class Gaussian(Distribution):
    """A Gaussian distribution.

    This represents a bell-curve centred at ``mean`` and with
    spread represented by the standard deviation, ``std``.

    Parameters
    ----------
    mean : Number
        The mean of the Gaussian.
    std : Number
        The standard deviation of the Gaussian.

    Raises
    ------
    ValidationError if std is <= 0

    """

    mean = NumberParam("mean")
    std = NumberParam("std", low=0, low_open=True)

    def __init__(self, mean, std):
        super().__init__()
        self.mean = mean
        self.std = std

    def sample(self, n, d=None, rng=np.random):
        shape = self._sample_shape(n, d)
        return rng.normal(loc=self.mean, scale=self.std, size=shape)


class Exponential(Distribution):
    """An exponential distribution (optionally with high values clipped).

    If ``high`` is left to its default value of infinity, this is a standard
    exponential distribution. If ``high`` is set, then any sampled values at
    or above ``high`` will be clipped so they are slightly below ``high``.
    This is useful for thresholding.

    The probability distribution function (PDF) is given by::

               |  0                                 if x < shift
        p(x) = | 1/scale * exp(-(x - shift)/scale)  if x >= shift and x < high
               |  n                                 if x == high - eps
               |  0                                 if x >= high

    where ``n`` is such that the PDF integrates to one, and ``eps`` is an
    infinitesimally small number such that samples of ``x`` are strictly less
    than ``high`` (in practice, ``eps`` depends on floating point precision).

    Parameters
    ----------
    scale : float
        The scale parameter (inverse of the rate parameter lambda). Larger
        values make the distribution narrower (sharper peak).
    shift : float, optional
        Amount to shift the distribution by. There will be no values smaller
        than this shift when sampling from the distribution.
    high : float, optional
        All values larger than or equal to this value will be clipped to
        slightly less than this value.
    """

    scale = NumberParam("scale", low=0, low_open=True)
    shift = NumberParam("shift")
    high = NumberParam("high")

    def __init__(self, scale, shift=0.0, high=np.inf):
        super().__init__()
        self.scale = scale
        self.shift = shift
        self.high = high

    def sample(self, n, d=None, rng=np.random):
        shape = self._sample_shape(n, d)
        x = rng.exponential(self.scale, shape) + self.shift
        high = np.nextafter(self.high, np.asarray(-np.inf, dtype=x.dtype))
        return npext.clip(x, self.shift, high)


class UniformHypersphere(Distribution):
    """Uniform distribution on or in an n-dimensional unit hypersphere.

    Sample points are uniformly distributed across the volume (default) or
    surface of an n-dimensional unit hypersphere.

    Parameters
    ----------
    surface : bool, optional
        Whether sample points should be distributed uniformly
        over the surface of the hyperphere (True),
        or within the hypersphere (False).
    min_magnitude : Number, optional
        Lower bound on the returned vector magnitudes (such that they are in
        the range ``[min_magnitude, 1]``). Must be in the range [0, 1).
        Ignored if ``surface`` is ``True``.
    """

    surface = BoolParam("surface")
    min_magnitude = NumberParam("min_magnitude", low=0, high=1, high_open=True)

    def __init__(self, surface=False, min_magnitude=0):
        super().__init__()
        if surface and min_magnitude > 0:
            warnings.warn("min_magnitude ignored because surface is True")
        self.surface = surface
        self.min_magnitude = min_magnitude

    def sample(self, n, d=None, rng=np.random):
        if d is None or d < 1:  # check this, since other dists allow d = None
            raise ValidationError("Dimensions must be a positive integer", "d")

        samples = rng.randn(n, d)
        samples /= npext.norm(samples, axis=1, keepdims=True)

        if self.surface:
            return samples

        # Generate magnitudes for vectors from uniform distribution.
        # The (1 / d) exponent ensures that samples are uniformly distributed
        # in n-space and not all bunched up at the centre of the sphere.
        samples *= rng.uniform(low=self.min_magnitude ** d, high=1, size=(n, 1)) ** (
            1.0 / d
        )

        return samples


class QuasirandomSequence(Distribution):
    """Sequence for quasi Monte Carlo sampling the ``[0, 1]``-cube.

    This is similar to ``np.random.uniform(0, 1, size=(num, d))``, but with the
    additional property that each ``d``-dimensional point is uniformly scattered.

    While the sequence is defined deterministically, we introduce two stochastic
    elements to encourage heterogeneity in models using these sequences.
    First, we offset the start of the sequence by a random number between 0 and 1
    to ensure we don't oversample points aligned to the step size.
    Second, we shuffle the resulting sequence before returning to ensure we don't
    introduce correlations between parameters sampled from this distribution.

    This is based on the tutorial and code from [#]_.

    See Also
    --------
    ScatteredHypersphere

    References
    ----------
    .. [#] Martin Roberts. "The Unreasonable Effectiveness of Quasirandom Sequences."
       https://web.archive.org/web/20211029094035/http://extremelearning.com.au/unreasonable-effectiveness-of-quasirandom-sequences/

    Examples
    --------

    .. testcode::

       rd = nengo.dists.QuasirandomSequence().sample(10000, 2)
       plt.scatter(*rd.T, c=np.arange(len(rd)), cmap='Blues', s=7)
    """

    @staticmethod
    def _phi(d, tol=1e-10):
        """Newton-Raphson-Method to calculate ``g = phi_d``."""
        x = 1.0
        for _ in range(100):
            dx = (x ** (d + 1) - x - 1) / ((d + 1) * x ** d - 1)
            x -= dx
            if abs(dx) < tol:
                break
        else:
            raise ConvergenceError(f"'phi' computation did not converge for d={d}")

        return x

    def sample(self, n, d=1, rng=np.random):
        if d is None or d < 1:  # check this, since other dists allow d = None
            raise ValidationError("Dimensions must be a positive integer", "d")

        seed = rng.uniform(0, 1, size=d)
        if d == 1:
            # Tile the points optimally
            z = (seed + (1 / n) * np.arange(n, dtype=np.float64)) % 1
            z = z[:, np.newaxis]
        else:
            phi_inv = 1.0 / self._phi(d)
            alpha = phi_inv ** np.arange(1, d + 1)
            z = (seed + np.outer(np.arange(1, n + 1), alpha)) % 1
        rng.shuffle(z)
        return z


class ScatteredHypersphere(Distribution):
    r"""Quasirandom distribution over the hypersphere or hyperball.

    Applies a spherical transform to the given quasirandom sequence
    (by default `.QuasirandomSequence`) to obtain uniformly scattered samples.

    This distribution has the nice mathematical property that the discrepancy
    between the empirical distribution and :math:`n` samples is
    :math:`\widetilde{\mathcal{O}} (1 / n)` as opposed to
    :math:`\mathcal{O} (1 / \sqrt{n})` for the Monte Carlo method [1]_.
    This means that the number of samples is effectively squared, making this
    useful as a means for sampling ``eval_points`` and ``encoders``.

    Parameters
    ----------
    surface : bool, optional
        Whether sample points should be distributed uniformly
        over the surface of the hyperphere (True),
        or within the hypersphere (False).
    min_magnitude : Number, optional
        Lower bound on the returned vector magnitudes (such that they are in
        the range ``[min_magnitude, 1]``). Must be in the range [0, 1).
        Ignored if ``surface`` is ``True``.
    base : `.Distribution`, optional
        The base distribution from which to sample quasirandom numbers.
    method : {"sct-approx", "sct", "tfww"}
        Method to use for mapping points to the hypersphere.

        * "sct-approx": Same as "sct", but uses lookup table to approximate the
          beta distribution, making it faster with almost exactly the same result.
        * "sct": Use the exact Spherical Coordinate Transform
          (section 1.5.2 of [1]_).
        * "tfww": Use the Tashiro-Fang-Wang-Wong method (section 4.3 of [1]_).
          Faster than "sct" and "sct-approx", with the same level of uniformity
          for larger numbers of samples (``n >= 4000``, approximately).

    See Also
    --------
    UniformHypersphere
    QuasirandomSequence

    Notes
    -----
    The `.QuasirandomSequence` distribution is mostly deterministic.
    Nondeterminism comes from a random ``d``-dimensional rotation.

    References
    ----------
    .. [1] K.-T. Fang and Y. Wang, Number-Theoretic Methods in Statistics.
       Chapman & Hall, 1994.

    Examples
    --------
    Plot points sampled from the surface of the sphere in 3 dimensions:

    .. testcode::

       from mpl_toolkits.mplot3d import Axes3D

       points = nengo.dists.ScatteredHypersphere(surface=True).sample(1000, d=3)

       ax = plt.subplot(111, projection="3d")
       ax.scatter(*points.T, s=5)

    Plot points sampled from the volume of the sphere in 2 dimensions (i.e. circle):

    .. testcode::

       points = nengo.dists.ScatteredHypersphere(surface=False).sample(1000, d=2)
       plt.scatter(*points.T, s=5)
    """

    surface = BoolParam("surface")
    min_magnitude = NumberParam("min_magnitude", low=0, high=1, high_open=True)
    base = DistributionParam("base")
    method = EnumParam("method", values=("sct-approx", "sct", "tfww"))

    def __init__(
        self,
        surface=False,
        min_magnitude=0,
        base=QuasirandomSequence(),
        method="sct-approx",
    ):
        super().__init__()
        if surface and min_magnitude > 0:
            warnings.warn("min_magnitude ignored because surface is True")
        self.surface = surface
        self.min_magnitude = min_magnitude
        self.base = base
        self.method = method

        if self.method == "sct":
            import scipy.special  # pylint: disable=import-outside-toplevel

            assert scipy.special

    @classmethod
    def spherical_coords_ppf(cls, dims, y, approx=False):
        if not approx:
            import scipy.special  # pylint: disable=import-outside-toplevel

        y_reflect = np.where(y < 0.5, y, 1 - y)
        if approx:
            z_sq = _betaincinv22.lookup(dims, 2 * y_reflect)
        else:
            z_sq = scipy.special.betaincinv(dims / 2.0, 0.5, 2 * y_reflect)
        x = np.arcsin(np.sqrt(z_sq)) / np.pi
        return np.where(y < 0.5, x, 1 - x)

    @classmethod
    def spherical_transform_sct(cls, samples, approx=False):
        """Map samples from the ``[0, 1]``-cube onto the hypersphere.

        Uses the SCT method described in section 1.5.3 of Fang and Wang (1994).
        """
        samples = np.asarray(samples)
        samples = samples[:, np.newaxis] if samples.ndim == 1 else samples
        n, d = samples.shape

        # inverse transform method (section 1.5.2)
        coords = np.empty_like(samples)
        for j in range(d):
            coords[:, j] = cls.spherical_coords_ppf(d - j, samples[:, j], approx=approx)

        # spherical coordinate transform
        mapped = np.ones((n, d + 1))
        i = np.ones(d)
        i[-1] = 2.0
        s = np.sin(i[np.newaxis, :] * np.pi * coords)
        c = np.cos(i[np.newaxis, :] * np.pi * coords)
        mapped[:, 1:] = np.cumprod(s, axis=1)
        mapped[:, :-1] *= c
        return mapped

    @staticmethod
    def spherical_transform_tfww(c_samples):
        """Map samples from the ``[0, 1]``-cube onto the hypersphere surface.

        Uses the TFWW method described in section 4.3 of Fang and Wang (1994).
        """
        c_samples = np.asarray(c_samples)
        c_samples = c_samples[:, np.newaxis] if c_samples.ndim == 1 else c_samples
        n, s1 = c_samples.shape
        s = s1 + 1

        x_samples = np.zeros((n, s))

        if s == 2:
            phi = 2 * np.pi * c_samples[:, 0]
            x_samples[:, 0] = np.cos(phi)
            x_samples[:, 1] = np.sin(phi)
            return x_samples

        even = s % 2 == 0
        m = s // 2 if even else (s - 1) // 2

        g = np.zeros((n, m + 1))
        g[:, -1] = 1
        for j in range(m - 1, 0, -1):
            g[:, j] = g[:, j + 1] * c_samples[:, j - 1] ** (
                (1.0 / j) if even else (2.0 / (2 * j + 1))
            )

        d = np.sqrt(np.diff(g, axis=1))

        phi = c_samples[:, m - 1 :]
        if even:
            phi *= 2 * np.pi
            x_samples[:, 0::2] = d * np.cos(phi)
            x_samples[:, 1::2] = d * np.sin(phi)
        else:
            # there is a mistake in eq. 4.3.7 here, see eq. 1.5.28 for correct version
            phi[:, 1:] *= 2 * np.pi
            f = 2 * d[:, 0] * np.sqrt(phi[:, 0] * (1 - phi[:, 0]))
            x_samples[:, 0] = d[:, 0] * (1 - 2 * phi[:, 0])
            x_samples[:, 1] = f * np.cos(phi[:, 1])
            x_samples[:, 2] = f * np.sin(phi[:, 1])
            if s > 3:
                x_samples[:, 3::2] = d[:, 1:] * np.cos(phi[:, 2:])
                x_samples[:, 4::2] = d[:, 1:] * np.sin(phi[:, 2:])

        return x_samples

    @staticmethod
    def random_orthogonal(d, rng=np.random):
        """Returns a random orthogonal matrix."""
        m = rng.standard_normal((d, d))
        u, _, v = np.linalg.svd(m)
        return np.dot(u, v)

    def sample(self, n, d=1, rng=np.random):
        if d == 1 and self.surface:
            return np.sign(self.base.sample(n, d, rng) - 0.5)

        if d == 1:
            pos_samples = self.base.sample(int(n / 2), d, rng)
            neg_samples = self.base.sample(n - pos_samples.size, d, rng)
            if self.min_magnitude > 0:
                for samples in [pos_samples, neg_samples]:
                    samples *= 1.0 - self.min_magnitude
                    samples += self.min_magnitude
            samples = np.vstack([pos_samples, -1 * neg_samples])
            rng.shuffle(samples)
            return samples

        radius = None
        if self.surface:
            samples = self.base.sample(n, d - 1, rng)
        else:
            samples = self.base.sample(n, d, rng)
            samples, radius = samples[:, :-1], samples[:, -1:]
            if self.min_magnitude != 0:
                min_d = self.min_magnitude ** d
                radius *= 1 - min_d
                radius += min_d
            radius **= 1.0 / d

        if self.method == "sct":
            mapped = self.spherical_transform_sct(samples, approx=False)
        elif self.method == "sct-approx":
            mapped = self.spherical_transform_sct(samples, approx=True)
        else:
            assert self.method == "tfww"
            mapped = self.spherical_transform_tfww(samples)

        # radius adjustment for ball
        if radius is not None:
            mapped *= radius

        # random rotation
        rotation = self.random_orthogonal(d, rng=rng)
        return np.dot(mapped, rotation)


class Choice(Distribution):
    """Discrete distribution across a set of possible values.

    The same as Numpy random's `~numpy.random.RandomState.choice`,
    except can take vector or matrix values for the choices.

    Parameters
    ----------
    options : (N, ...) array_like
        The options (choices) to choose between. The choice is always done
        along the first axis, so if ``options`` is a matrix, the options are
        the rows of that matrix.
    weights : (N,) array_like, optional
        Weights controlling the probability of selecting each option. Will
        automatically be normalized. If None, weights be uniformly distributed.
    """

    options = NdarrayParam("options", shape=("*", "..."))
    weights = NdarrayParam("weights", shape=("*"), optional=True)

    def __init__(self, options, weights=None):
        super().__init__()
        self.options = options
        self.weights = weights

        weights = np.ones(len(self.options)) if self.weights is None else self.weights
        if len(weights) != len(self.options):
            raise ValidationError(
                f"Number of weights ({len(weights)}) must match "
                f"number of options ({len(self.options)})",
                attr="weights",
                obj=self,
            )
        if not all(weights >= 0):
            raise ValidationError(
                "All weights must be non-negative", attr="weights", obj=self
            )
        total = float(weights.sum())
        if total <= 0:
            raise ValidationError(
                f"Sum of weights must be positive (got {total:f})",
                attr="weights",
                obj=self,
            )
        self.p = weights / total

    @property
    def dimensions(self):
        return 0 if self.options.ndim == 1 else np.prod(self.options.shape[1:])

    def sample(self, n, d=None, rng=np.random):
        if d is not None and self.dimensions != d:
            raise ValidationError(
                f"Options must be of dimensionality {d} (got {self.dimensions})",
                attr="options",
                obj=self,
            )

        i = np.searchsorted(np.cumsum(self.p), rng.rand(n))
        return self.options[i]


class Samples(Distribution):
    """A set of samples.

    This class is a subclass of `.Distribution` so that it can be used in any
    situation that calls for a  `.Distribution`. However, the call to
    `.Distribution.sample` must match the dimensions of the samples or
    a `.ValidationError` will be raised.

    Parameters
    ----------
    samples : (n, d) array_like
        ``n`` and ``d`` must match what is eventually passed to
         `.Distribution.sample`.
    """

    samples = NdarrayParam("samples", shape=("...",))

    def __init__(self, samples):
        super().__init__()
        self.samples = samples

    def sample(self, n, d=None, rng=np.random):
        samples = np.array(self.samples)
        shape = (n,) if d is None else (n, d)

        if d is None:
            samples = samples.squeeze()

        if d is not None and samples.ndim == 1:
            samples = samples[..., np.newaxis]

        if samples.shape[0] != shape[0]:
            raise ValidationError(
                "Wrong number of samples requested; got "
                f"{n}, should be {samples.shape[0]}",
                attr="samples",
                obj=self,
            )
        elif d is None and len(samples.shape) != 1:
            raise ValidationError(
                "Wrong sample dimensionality requested; got "
                f"'None', should be {samples.shape[1]}",
                attr="samples",
                obj=self,
            )
        elif d is not None and samples.shape[1] != shape[1]:
            raise ValidationError(
                "Wrong sample dimensionality requested; got "
                f"{d}, should be {samples.shape[1]}",
                attr="samples",
                obj=self,
            )

        return samples


class SqrtBeta(Distribution):
    """Distribution of the square root of a Beta distributed random variable.

    Given ``n + m`` dimensional random unit vectors, the length of subvectors
    with ``m`` elements will be distributed according to this distribution.

    Parameters
    ----------
    n: int
        Number of subvectors.
    m: int, optional
        Length of each subvector.

    See also
    --------
    nengo.dists.SubvectorLength
    """

    n = IntParam("n", low=0)
    m = IntParam("m", low=0)

    def __init__(self, n, m=1):
        super().__init__()
        self.n = n
        self.m = m

    def sample(self, n, d=None, rng=np.random):
        shape = self._sample_shape(n, d)
        return np.sqrt(rng.beta(self.m / 2.0, self.n / 2.0, size=shape))

    def cdf(self, x):
        """Cumulative distribution function.

        .. note:: Requires SciPy.

        Parameters
        ----------
        x : array_like
            Evaluation points in [0, 1].

        Returns
        -------
        cdf : array_like
            Probability that ``X <= x``.
        """
        from scipy.special import betainc  # pylint: disable=import-outside-toplevel

        sq_x = x * x
        return np.where(
            sq_x < 1.0, betainc(self.m / 2.0, self.n / 2.0, sq_x), np.ones_like(x)
        )

    def pdf(self, x):
        """Probability distribution function.

        .. note:: Requires SciPy.

        Parameters
        ----------
        x : array_like
            Evaluation points in [0, 1].

        Returns
        -------
        pdf : array_like
            Probability density at ``x``.
        """
        from scipy.special import beta  # pylint: disable=import-outside-toplevel

        return (
            2
            / beta(0.5 * self.m, 0.5 * self.n)
            * x ** (self.m - 1)
            * (1 - x * x) ** (0.5 * self.n - 1)
        )

    def ppf(self, y):
        """Percent point function (inverse cumulative distribution).

        .. note:: Requires SciPy.

        Parameters
        ----------
        y : array_like
            Cumulative probabilities in [0, 1].

        Returns
        -------
        ppf : array_like
            Evaluation points ``x`` in [0, 1] such that ``P(X <= x) = y``.
        """
        from scipy.special import betaincinv  # pylint: disable=import-outside-toplevel

        sq_x = betaincinv(self.m / 2.0, self.n / 2.0, y)
        return np.sqrt(sq_x)


class SubvectorLength(SqrtBeta):
    """Distribution of the length of a subvectors of a unit vector.

    Parameters
    ----------
    dimensions : int
        Dimensionality of the complete unit vector.
    subdimensions : int, optional
        Dimensionality of the subvector.

    See also
    --------
    nengo.dists.SqrtBeta
    """

    def __init__(self, dimensions, subdimensions=1):
        super().__init__(dimensions - subdimensions, subdimensions)

    @property
    def dimensions(self):
        return self.n + self.m

    @property
    def subdimensions(self):
        return self.m


class CosineSimilarity(SubvectorLength):
    """Distribution of the cosine of the angle between two random vectors.

    The "cosine similarity" is the cosine of the angle between two vectors,
    which is equal to the dot product of the vectors, divided by the L2-norms
    of the individual vectors. When these vectors are unit length, this is then
    simply the distribution of their dot product.

    This is also equivalent to the distribution of a single coefficient from a
    unit vector (a single dimension of ``UniformHypersphere(surface=True)``).
    Furthermore, ``CosineSimilarity(d+2)`` is equivalent to the distribution of
    a single coordinate from points uniformly sampled from the d-dimensional
    unit ball (a single dimension of
    ``UniformHypersphere(surface=False).sample(n, d)``). These relationships
    have been detailed in [Voelker2017]_.

    This can be used to calculate an intercept ``c = ppf(1 - p)`` such that
    ``dot(u, v) >= c`` with probability ``p``, for random unit vectors ``u``
    and ``v``. In other words, a neuron with intercept ``ppf(1 - p)`` will
    fire with probability ``p`` for a random unit length input.

    .. [Voelker2017]
       `Aaron R. Voelker, Jan Gosmann, and Terrence C. Stewart.
       Efficiently sampling vectors and coordinates from the n-sphere and
       n-ball. Technical Report, Centre for Theoretical Neuroscience,
       Waterloo, ON, 2017
       <http://compneuro.uwaterloo.ca/publications/voelker2017.html>`_

    Parameters
    ----------
    dimensions: int
        Dimensionality of the complete unit vector.

    See also
    --------
    nengo.dists.SqrtBeta
    """

    def __init__(self, dimensions):
        super().__init__(dimensions)

    def sample(self, n, d=None, rng=np.random):
        shape = self._sample_shape(n, d)
        sign = Choice((1, -1)).sample(np.prod(shape), rng=rng).reshape(*shape)
        return sign * super().sample(n, d, rng=rng)

    def cdf(self, x):
        return (super().cdf(x) * np.sign(x) + 1) / 2.0

    def pdf(self, x):
        return super().pdf(x) / 2.0

    def ppf(self, y):
        x = super().ppf(abs(y * 2 - 1))
        return np.where(y > 0.5, x, -x)


class _betaincinv22:
    """Look up values for ``betaincinv(dims / 2, 0.5, x)``."""

    path = str(data_dir / "betaincinv22_table.npz")
    table = None

    @classmethod
    def make_table(cls, n_interp=200, n_dims=50):
        """Save lookup table for ``_betaincinv22.lookup``."""
        import scipy.special  # pylint: disable=import-outside-toplevel

        rng = np.random.RandomState(0)

        n_dims_log = int(0.5 * n_dims)
        dims_lin = np.arange(1, n_dims - n_dims_log + 1)
        dims_log = np.round(
            np.logspace(np.log10(dims_lin[-1] + 1), 3, n_dims_log)
        ).astype(dims_lin.dtype)
        dims = np.unique(np.concatenate([dims_lin, dims_log]))

        x_table = []
        y_table = []
        for dim in dims:
            n_range = int(0.8 * n_interp)
            x0 = np.linspace(0, 1, n_interp - n_range)  # samples in the domain
            y0 = np.linspace(1e-16, 1 - 1e-7, n_range)  # samples in the range
            x1 = scipy.special.betainc(dim / 2.0, 0.5, y0)
            interp_x = np.unique(np.concatenate([x0, x1]))
            while len(interp_x) < n_interp:
                # add random points until we achieve the length
                interp_x = np.unique(
                    np.concatenate(
                        [interp_x, rng.uniform(size=n_interp - len(interp_x))]
                    )
                )

            interp_x.sort()
            assert interp_x.size == n_interp

            interp_y = scipy.special.betaincinv(dim / 2.0, 0.5, interp_x)
            x_table.append(interp_x)
            y_table.append(interp_y)

        x_table = np.asarray(x_table)
        y_table = np.asarray(y_table)

        np.savez(cls.path, dims=dims, x=x_table, y=y_table)

    @classmethod
    def load_table(cls):
        if cls.table is not None:
            return cls.table
        data = np.load(cls.path)
        cls.table = {d: (x, y) for d, x, y in zip(data["dims"], data["x"], data["y"])}
        assert np.all(np.diff(list(cls.table)) >= 1)
        return cls.table

    @classmethod
    def lookup(cls, dims, x):
        if not is_integer(dims) or dims < 1:
            raise ValidationError("must be an integer >= 1", attr="dims")

        table = cls.load_table()

        if dims in table:  # pylint: disable=unsupported-membership-test
            xp, yp = table[dims]
        else:
            known_dims = np.array(list(table))
            i = np.searchsorted(known_dims, dims)
            assert i > 0
            if i >= len(known_dims):
                # dims is larger than any dimension we have, so use the largest
                xp, yp = table[known_dims[-1]]
            else:
                # take average of two curves
                dims0, dims1 = known_dims[i - 1], known_dims[i]
                xp0, yp0 = table[dims0]
                xp1, yp1 = table[dims1]
                assert dims0 < dims < dims1
                ratio0 = (dims1 - dims) / (dims1 - dims0)
                ratio1 = 1 - ratio0
                xp = (ratio0 * xp0 + ratio1 * xp1) if len(xp0) == len(xp1) else xp0
                yp = ratio0 * np.interp(xp, xp0, yp0) + ratio1 * np.interp(xp, xp1, yp1)

        return np.interp(x, xp, yp)



import numpy as np

import nengo


def test_basic(Simulator, plt, seed):
    with nengo.Network(seed=seed) as net:
        bg = nengo.networks.BasalGanglia(dimensions=5)
        input = nengo.Node([0.8, 0.4, 0.4, 0.4, 0.4], label="input")
        nengo.Connection(input, bg.input, synapse=None)
        p = nengo.Probe(bg.output, synapse=0.01)

    with Simulator(net) as sim:
        sim.run(0.2)

    t = sim.trange()
    output = np.mean(sim.data[p][t > 0.1], axis=0)

    plt.plot(t, sim.data[p])
    plt.ylabel("Output")

    assert output[0] > -0.1
    assert np.all(output[1:] < -0.8)


def test_thalamus(Simulator, plt, seed):
    with nengo.Network(seed=seed) as net:
        bg = nengo.networks.BasalGanglia(dimensions=5)
        input = nengo.Node([0.8, 0.4, 0.4, 0.4, 0.4], label="input")
        nengo.Connection(input, bg.input, synapse=None)

        thal = nengo.networks.Thalamus(dimensions=5)
        nengo.Connection(bg.output, thal.input)

        p = nengo.Probe(thal.output, synapse=0.01)

    with Simulator(net) as sim:
        sim.run(0.2)

    t = sim.trange()
    output = np.mean(sim.data[p][t > 0.1], axis=0)

    plt.plot(t, sim.data[p])
    plt.ylabel("Output")

    assert output[0] > 0.8
    assert np.all(output[1:] < 0.01)


def test_bias_input(Simulator, plt, seed, allclose):
    with nengo.Network(seed=seed) as net:
        bg = nengo.networks.BasalGanglia(dimensions=3, input_bias=-0.5)
        input = nengo.Node([0.8, 0.5, 0.2], label="input")
        nengo.Connection(input, bg.input, synapse=None)
        p_in = nengo.Probe(bg.input, synapse=0.01)

    with Simulator(net) as sim:
        sim.run(0.1)

    t = sim.trange()
    plt.plot(t, sim.data[p_in])
    plt.ylabel("Input")

    assert allclose(sim.data[p_in][t > 0.08], [0.3, 0, -0.3], atol=0.005)


def test_overridden_configs():
    # Using default synapses, all that are not None should be
    # Lowpass with tau of 0.002 (AMPA) or 0.008 (GABA)
    bg = nengo.networks.BasalGanglia(dimensions=2)
    for connection in bg.all_connections:
        if connection.synapse is None:
            continue
        assert isinstance(connection.synapse, nengo.Lowpass)
        assert connection.synapse.tau in (0.002, 0.008)

    # These configs do not override the synapse on the connection.
    # The synapses should be the same as before, and the config should
    # be unchanged after instantiation, even though inside the class
    # it is changed.
    ampa_config = nengo.Config(nengo.Connection)
    ampa_config[nengo.Connection].solver = nengo.solvers.LstsqL2nz()
    assert "synapse" not in ampa_config[nengo.Connection]
    gaba_config = nengo.Config(nengo.Connection)
    gaba_config[nengo.Connection].solver = nengo.solvers.LstsqL2nz()
    assert "synapse" not in gaba_config[nengo.Connection]

    bg = nengo.networks.BasalGanglia(
        dimensions=2, ampa_config=ampa_config, gaba_config=gaba_config
    )
    for connection in bg.all_connections:
        if connection.synapse is None:
            continue
        # Same synapses
        assert isinstance(connection.synapse, nengo.Lowpass)
        assert connection.synapse.tau in (0.002, 0.008)
        # But solver should be different
        assert isinstance(connection.solver, nengo.solvers.LstsqL2nz)
    # Ensure default synapse has been removed from both objects
    assert "synapse" not in ampa_config[nengo.Connection]
    assert "synapse" not in gaba_config[nengo.Connection]

    # This config overrides the synapse on AMPA and GABA connections
    config_with_synapse = nengo.Config(nengo.Connection)
    config_with_synapse[nengo.Connection].synapse = nengo.Alpha(0.1)
    bg = nengo.networks.BasalGanglia(
        dimensions=2, ampa_config=config_with_synapse, gaba_config=config_with_synapse
    )
    for connection in bg.all_connections:
        if connection.synapse is None:
            continue
        # Synapse changes
        assert isinstance(connection.synapse, nengo.Alpha)
        assert connection.synapse.tau == 0.1
        # Solver should be the same
        assert not isinstance(connection.solver, nengo.solvers.LstsqL1)
    # Ensure synapse is the same as before passing to BasalGanglia
    synapse = config_with_synapse[nengo.Connection].synapse
    assert isinstance(synapse, nengo.Alpha)
    assert synapse.tau == 0.1



from nengo.builder.learning_rules import SimBCM, SimOja, SimPES, SimVoja
from nengo.builder.tests.test_operator import _test_operator_arg_attributes


def test_simpes():
    argnames = ["pre_filtered", "error", "delta", "learning_rate"]
    non_signals = ["learning_rate"]
    _, sim = _test_operator_arg_attributes(SimPES, argnames, non_signals=non_signals)

    assert str(sim) == "SimPES{pre=pre_filtered, error=error -> delta}"


def test_simbcm():
    argnames = ["pre_filtered", "post_filtered", "theta", "delta", "learning_rate"]
    non_signals = ["learning_rate"]
    _, sim = _test_operator_arg_attributes(SimBCM, argnames, non_signals=non_signals)

    assert str(sim) == "SimBCM{pre=pre_filtered, post=post_filtered -> delta}"


def test_simoja():
    argnames = [
        "pre_filtered",
        "post_filtered",
        "weights",
        "delta",
        "learning_rate",
        "beta",
    ]
    non_signals = ["learning_rate", "beta"]
    _, sim = _test_operator_arg_attributes(SimOja, argnames, non_signals=non_signals)

    assert str(sim) == "SimOja{pre=pre_filtered, post=post_filtered -> delta}"


def test_simvoja():
    argnames = [
        "pre_decoded",
        "post_filtered",
        "scaled_encoders",
        "delta",
        "scale",
        "learning_signal",
        "learning_rate",
    ]
    non_signals = ["learning_rate", "scale"]
    args, sim = _test_operator_arg_attributes(
        SimVoja, argnames, non_signals=non_signals
    )
    assert sim.weights is args["scaled_encoders"]

    assert str(sim) == "SimVoja{pre=pre_decoded, post=post_filtered -> delta}"



import numpy as np
import pytest

import nengo
from nengo.utils.matplotlib import rasterplot, set_color_cycle


@pytest.mark.parametrize("use_eventplot", [True, False])
def test_rasterplot(use_eventplot, Simulator, seed, plt):
    with nengo.Network(seed=seed) as model:
        # use amplitude 0.5 to have some neurons not spike
        u = nengo.Node(output=lambda t: 0.5 * np.sin(6 * t))
        a = nengo.Ensemble(100, 1)
        nengo.Connection(u, a)
        ap = nengo.Probe(a.neurons)

    with Simulator(model) as sim:
        sim.run(1.0)

    rasterplot(sim.trange(), sim.data[ap], use_eventplot=use_eventplot)

    # TODO: add assertions


def test_rasterplot_onetime(plt):
    """Tests rasterplot with a single time point"""
    time = [0.2]
    spikes = np.array([1, 0, 0, 1, 0, 1, 1]).reshape((1, -1))
    rasterplot(time, spikes, ax=None)


def test_set_color_cycle(plt):
    if plt.__file__ == "/dev/null":
        pytest.skip("Only runs when plotting is enabled")

    slopes = np.array([-0.4, 0.3, 0.5])
    t = np.linspace(0, 1)
    y = t[:, None] * slopes

    colors1 = ["green", "blue", "red"]
    colors2 = ["red", "green", "blue"]

    # set ax1 colors with `ax` arg, and ax2 colors with `ax=None`
    ax1 = plt.subplot(211)
    set_color_cycle(colors2, ax=None)  # must be done BEFORE the axis is created
    ax2 = plt.subplot(212)

    for i in range(slopes.size):
        ax2.plot(y[:, i], label=colors2[i])
    plt.legend()

    set_color_cycle(colors1, ax=ax1)
    for i in range(slopes.size):
        ax1.plot(y[:, i], label=colors1[i])
    ax1.legend()

    for i in range(slopes.size):
        assert ax1.lines[i].get_color() == colors1[i]
        assert ax2.lines[i].get_color() == colors2[i]



import warnings

import numpy as np

from nengo.base import FrozenObject
from nengo.dists import DistOrArrayParam, Distribution, Uniform
from nengo.exceptions import ValidationError
from nengo.params import (
    BoolParam,
    EnumParam,
    IntParam,
    NdarrayParam,
    Parameter,
    ShapeParam,
)
from nengo.rc import rc
from nengo.utils.numpy import is_array_like, scipy_sparse


class Transform(FrozenObject):
    """A base class for connection transforms.

    .. versionadded:: 3.0.0
    """

    def sample(self, rng=np.random):
        """Returns concrete weights to implement the specified transform.

        Parameters
        ----------
        rng : `numpy.random.RandomState`, optional
            Random number generator state.

        Returns
        -------
        array_like
            Transform weights
        """
        raise NotImplementedError()

    @property
    def size_in(self):
        """Expected size of input to transform."""
        raise NotImplementedError()

    @property
    def size_out(self):
        """Expected size of output from transform."""
        raise NotImplementedError()


class ChannelShapeParam(ShapeParam):
    """A parameter where the value must be a shape with channels.

    .. versionadded:: 3.0.0
    """

    def coerce(self, transform, shape):  # pylint: disable=arguments-renamed
        if isinstance(shape, ChannelShape):
            if shape.channels_last != transform.channels_last:
                raise ValidationError(
                    f"transform has channels_last={transform.channels_last}, but input "
                    f"shape has channels_last={shape.channels_last}",
                    attr=self.name,
                    obj=transform,
                )
            super().coerce(transform, shape.shape)
        else:
            super().coerce(transform, shape)
            shape = ChannelShape(shape, channels_last=transform.channels_last)
        return shape


class Dense(Transform):
    """A dense matrix transformation between an input and output signal.

    .. versionadded:: 3.0.0

    Parameters
    ----------
    shape : tuple of int
        The shape of the dense matrix: ``(size_out, size_in)``.
    init : `.Distribution` or array_like, optional
        A Distribution used to initialize the transform matrix, or a concrete
        instantiation for the matrix. If the matrix is square we also allow a
        scalar (equivalent to ``np.eye(n) * init``) or a vector (equivalent to
        ``np.diag(init)``) to represent the matrix more compactly.
    """

    shape = ShapeParam("shape", length=2, low=1)
    init = DistOrArrayParam("init")

    def __init__(self, shape, init=1.0):
        super().__init__()

        self.shape = shape

        if is_array_like(init):
            init = np.asarray(init, dtype=rc.float_dtype)

            # check that the shape of init is compatible with the given shape
            # for this transform
            expected_shape = None
            if shape[0] != shape[1]:
                # init must be 2D if transform is not square
                expected_shape = shape
            elif init.ndim == 1:
                expected_shape = (shape[0],)
            elif init.ndim >= 2:
                expected_shape = shape

            if expected_shape is not None and init.shape != expected_shape:
                raise ValidationError(
                    f"Shape of initial value {init.shape} does not match expected "
                    f"shape {expected_shape}",
                    attr="init",
                )

        self.init = init

    @property
    def _argreprs(self):
        return [f"shape={self.shape!r}"]

    def sample(self, rng=np.random):
        if isinstance(self.init, Distribution):
            return self.init.sample(*self.shape, rng=rng)

        return self.init

    @property
    def init_shape(self):
        """The shape of the initial value."""
        return self.shape if isinstance(self.init, Distribution) else self.init.shape

    @property
    def size_in(self):
        return self.shape[1]

    @property
    def size_out(self):
        return self.shape[0]


class SparseInitParam(Parameter):
    def coerce(self, instance, value):
        if not (
            isinstance(value, SparseMatrix)
            or (scipy_sparse is not None and isinstance(value, scipy_sparse.spmatrix))
        ):
            raise ValidationError(
                "Must be `nengo.transforms.SparseMatrix` or "
                f"`scipy.sparse.spmatrix`, got '{type(value)}'",
                attr="init",
                obj=instance,
            )
        return super().coerce(instance, value)


class SparseMatrix(FrozenObject):
    """Represents a sparse matrix.

    .. versionadded:: 3.0.0

    Parameters
    ----------
    indices : array_like of int
        An Nx2 array of integers indicating the (row,col) coordinates for the
        N non-zero elements in the matrix.
    data : array_like or `.Distribution`
        An Nx1 array defining the value of the nonzero elements in the matrix
        (corresponding to ``indices``), or a `.Distribution` that will be
        used to initialize the nonzero elements.
    shape : tuple of int
        Shape of the full matrix.
    """

    indices = NdarrayParam("indices", shape=("*", 2), dtype=np.int64)
    data = DistOrArrayParam("data", sample_shape=("*",))
    shape = ShapeParam("shape", length=2)

    def __init__(self, indices, data, shape):
        super().__init__()

        self.indices = indices
        self.shape = shape

        # if data is not a distribution
        if is_array_like(data):
            data = np.asarray(data)

            # convert scalars to vectors
            if data.size == 1:
                data = data.item() * np.ones(self.indices.shape[0], dtype=data.dtype)

            if data.ndim != 1 or data.shape[0] != self.indices.shape[0]:
                raise ValidationError(
                    "Must be a vector of the same length as `indices`",
                    attr="data",
                    obj=self,
                )

        self.data = data
        self._allocated = None
        self._dense = None

    @property
    def dtype(self):
        return self.data.dtype

    @property
    def ndim(self):
        return len(self.shape)

    @property
    def size(self):
        return self.indices.shape[0]

    def allocate(self):
        """Return a `scipy.sparse.csr_matrix` or dense matrix equivalent.

        We mark this data as readonly to be consistent with how other
        data associated with signals are allocated. If this allocated
        data is to be modified, it should be copied first.
        """

        if self._allocated is not None:
            return self._allocated

        if scipy_sparse is None:
            warnings.warn(
                "Sparse operations require Scipy, which is not "
                "installed. Using dense matrices instead."
            )
            self._allocated = self.toarray().view()
        else:
            self._allocated = scipy_sparse.csr_matrix(
                (self.data, self.indices.T), shape=self.shape
            )
            self._allocated.data.setflags(write=False)

        return self._allocated

    def sample(self, rng=np.random):
        """Convert `.Distribution` data to fixed array.

        Parameters
        ----------
        rng : `.numpy.random.RandomState`
            Random number generator that will be used when
            sampling distribution.

        Returns
        -------
        matrix : `.SparseMatrix`
            A new `.SparseMatrix` instance with `.Distribution` converted to
            array if ``self.data`` is a `.Distribution`, otherwise simply
            returns ``self``.
        """
        if isinstance(self.data, Distribution):
            return SparseMatrix(
                self.indices,
                self.data.sample(self.indices.shape[0], rng=rng),
                self.shape,
            )
        else:
            return self

    def toarray(self):
        """Return the dense matrix equivalent of this matrix."""

        if self._dense is not None:
            return self._dense

        self._dense = np.zeros(self.shape, dtype=self.dtype)
        self._dense[self.indices[:, 0], self.indices[:, 1]] = self.data
        # Mark as readonly, if the user wants to modify they should copy first
        self._dense.setflags(write=False)
        return self._dense


class Sparse(Transform):
    """A sparse matrix transformation between an input and output signal.

    .. versionadded:: 3.0.0

    Parameters
    ----------
    shape : tuple of int
        The full shape of the sparse matrix: ``(size_out, size_in)``.
    indices : array_like of int
        An Nx2 array of integers indicating the (row,col) coordinates for the
        N non-zero elements in the matrix.
    init : `.Distribution` or array_like, optional
        A Distribution used to initialize the transform matrix, or a concrete
        instantiation for the matrix. If the matrix is square we also allow a
        scalar (equivalent to ``np.eye(n) * init``) or a vector (equivalent to
        ``np.diag(init)``) to represent the matrix more compactly.
    """

    shape = ShapeParam("shape", length=2, low=1)
    init = SparseInitParam("init")

    def __init__(self, shape, indices=None, init=1.0):
        super().__init__()

        self.shape = shape

        if scipy_sparse and isinstance(init, scipy_sparse.spmatrix):
            assert indices is None
            assert init.shape == self.shape
            self.init = init
        elif indices is not None:
            self.init = SparseMatrix(indices, init, shape)
        else:
            raise ValidationError(
                "Either `init` must be a `scipy.sparse.spmatrix`, "
                "or `indices` must be specified.",
                attr="init",
            )

    @property
    def _argreprs(self):
        return [f"shape={self.shape!r}"]

    def sample(self, rng=np.random):
        if scipy_sparse and isinstance(self.init, scipy_sparse.spmatrix):
            return self.init
        else:
            return self.init.sample(rng=rng)

    @property
    def size_in(self):
        return self.shape[1]

    @property
    def size_out(self):
        return self.shape[0]


class _ConvolutionBase(Transform):
    """Abstract base class for Convolution and ConvolutionTranspose."""

    n_filters = IntParam("n_filters", low=1)
    input_shape = ChannelShapeParam("input_shape", low=1)
    kernel_size = ShapeParam("kernel_size", low=1)
    strides = ShapeParam("strides", low=1)
    padding = EnumParam("padding", values=("same", "valid"))
    channels_last = BoolParam("channels_last")
    init = DistOrArrayParam("init")

    _param_init_order = ["channels_last", "input_shape"]

    def __init__(
        self,
        n_filters,
        input_shape,
        kernel_size=(3, 3),
        strides=(1, 1),
        padding="valid",
        channels_last=True,
        init=Uniform(-1, 1),
    ):
        super().__init__()

        self.n_filters = n_filters
        self.channels_last = channels_last  # must be set before input_shape
        self.input_shape = input_shape
        self.kernel_size = kernel_size
        self.strides = strides
        self.padding = padding
        self.init = init

        if len(kernel_size) != self.dimensions:
            raise ValidationError(
                f"Kernel dimensions ({len(kernel_size)}) does not match "
                f"input dimensions ({self.dimensions})",
                attr="kernel_size",
            )
        if len(strides) != self.dimensions:
            raise ValidationError(
                f"Stride dimensions ({len(strides)}) does not match "
                f"input dimensions ({self.dimensions})",
                attr="strides",
            )
        if not isinstance(init, Distribution):
            if init.shape != self.kernel_shape:
                raise ValidationError(
                    f"Kernel shape {init.shape} does not match "
                    f"expected shape {self.kernel_shape}",
                    attr="init",
                )

    @property
    def _argreprs(self):
        argreprs = [
            f"n_filters={self.n_filters!r}",
            f"input_shape={self.input_shape.shape}",
        ]
        if self.kernel_size != (3, 3):
            argreprs.append(f"kernel_size={self.kernel_size!r}")
        if self.strides != (1, 1):
            argreprs.append(f"strides={self.strides!r}")
        if self.padding != "valid":
            argreprs.append(f"padding={self.padding!r}")
        if self.channels_last is not True:
            argreprs.append(f"channels_last={self.channels_last!r}")
        return argreprs

    def sample(self, rng=np.random):
        if isinstance(self.init, Distribution):
            # we sample this way so that any variancescaling distribution based
            # on n/d is scaled appropriately
            kernel = [
                self.init.sample(self.input_shape.n_channels, self.n_filters, rng=rng)
                for _ in range(np.prod(self.kernel_size))
            ]
            kernel = np.reshape(kernel, self.kernel_shape)
        else:
            kernel = np.array(self.init, dtype=rc.float_dtype)
        return kernel

    @property
    def kernel_shape(self):
        """Full shape of kernel."""
        return self.kernel_size + (self.input_shape.n_channels, self.n_filters)

    @property
    def size_in(self):
        return self.input_shape.size

    @property
    def size_out(self):
        return self.output_shape.size

    @property
    def dimensions(self):
        """Dimensionality of convolution."""
        return self.input_shape.dimensions

    def _forward_shape(self, input_spatial_shape, n_filters):
        output_shape = np.array(input_spatial_shape, dtype=rc.float_dtype)
        if self.padding == "valid":
            output_shape -= self.kernel_size
            output_shape += 1
        output_shape /= self.strides
        output_shape = tuple(np.ceil(output_shape).astype(rc.int_dtype))

        return ChannelShape.from_space_and_channels(
            output_shape, n_filters, channels_last=self.channels_last
        )


class Convolution(_ConvolutionBase):
    """An N-dimensional convolutional transform.

    The dimensionality of the convolution is determined by the input shape.

    .. versionadded:: 3.0.0

    Parameters
    ----------
    n_filters : int
        The number of convolutional filters to apply.
    input_shape : tuple of int or `.ChannelShape`
        Shape of the input signal to the convolution; e.g.,
        ``(height, width, channels)`` for a 2D convolution with ``channels_last=True``.
    kernel_size : tuple of int, optional
        Size of the convolutional kernels (1 element for a 1D convolution,
        2 for a 2D convolution, etc.).
    strides : tuple of int, optional
        Stride of the convolution (1 element for a 1D convolution, 2 for
        a 2D convolution, etc.).
    padding : ``"same"`` or ``"valid"``, optional
        Padding method for input signal. "Valid" means no padding, and
        convolution will only be applied to the fully-overlapping areas of the
        input signal (meaning the output will be smaller). "Same" means that
        the input signal is zero-padded so that the output is the same shape
        as the input.
    channels_last : bool, optional
        If ``True`` (default), the channels are the last dimension in the input
        signal (e.g., a 28x28 image with 3 channels would have shape
        ``(28, 28, 3)``).  ``False`` means that channels are the first
        dimension (e.g., ``(3, 28, 28)``).
    init : `.Distribution` or `~numpy:numpy.ndarray`, optional
        A predefined kernel with shape ``kernel_size + (input_channels, n_filters)``,
        or a ``Distribution`` that will be used to initialize the kernel.

    Notes
    -----
    As is typical in neural networks, this is technically correlation rather
    than convolution (because the kernel is not flipped).
    """

    def __init__(
        self,
        n_filters,
        input_shape,
        kernel_size=(3, 3),
        strides=(1, 1),
        padding="valid",
        channels_last=True,
        init=Uniform(-1, 1),
    ):
        super().__init__(
            n_filters=n_filters,
            input_shape=input_shape,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            channels_last=channels_last,
            init=init,
        )

        if self.padding == "valid":
            for i in range(self.dimensions):
                if self.kernel_size[i] > self.input_shape.spatial_shape[i]:
                    raise ValidationError(
                        f"Kernel size for spatial dimension {i} "
                        f"({self.kernel_size[i]}) exceeds the spatial size of that "
                        f"dimension ({self.input_shape.spatial_shape[i]}). With the "
                        "requested 'valid' padding, this will result in empty output.",
                        attr="padding",
                        obj=self,
                    )

    @property
    def output_shape(self):
        """Output shape after applying convolution to input."""
        return self._forward_shape(self.input_shape.spatial_shape, self.n_filters)


class ConvolutionTranspose(_ConvolutionBase):
    """An N-dimensional transposed convolutional transform.

    This performs the transpose operation of `.Convolution`. The ``kernel_size``,
    ``strides``, and ``padding`` parameters all act as in `.Convolution`, so this
    transform will be the transpose of a `.Convolution` transform with those
    parameters. The ``n_filters`` and ``input_shape`` parameters are relative to *this*
    transform. The output shape is ambiguous, and can thus be specified (i.e. with
    `.Convolution`, there can be more than one input shape that produces the same output
    shape, so here, there are multiple valid output shapes for some input shapes).

    The dimensionality of the transpose convolution is determined by the input shape.

    .. versionadded:: 3.2.0

    Parameters
    ----------
    n_filters : int
        The number of channels in the *output* of this transform.
    input_shape : tuple of int or `.ChannelShape`
        Shape of the input signal to this transform; e.g.,
        ``(height, width, channels)`` for a 2D convolution with ``channels_last=True``.
    output_shape : tuple of int or `.ChannelShape`, optional
        Shape of the output signal of this transform; e.g.,
        ``(output_height, output_width, n_filters)`` for a 2D convolution with
        ``channels_last=True``. Defaults to the smallest valid output shape.
    kernel_size : tuple of int, optional
        Size of the convolutional kernels (1 element for a 1D convolution,
        2 for a 2D convolution, etc.).
    strides : tuple of int, optional
        Stride of the convolution (1 element for a 1D convolution, 2 for
        a 2D convolution, etc.).
    padding : ``"same"`` or ``"valid"``, optional
        Padding method for corresponding `.Convolution`.
    channels_last : bool, optional
        If ``True`` (default), the channels are the last dimension in the input
        signal (e.g., a 28x28 image with 3 channels would have shape
        ``(28, 28, 3)``).  ``False`` means that channels are the first
        dimension (e.g., ``(3, 28, 28)``).
    init : `.Distribution` or `~numpy:numpy.ndarray`, optional
        A predefined kernel with shape ``kernel_size + (input_channels, n_filters)``,
        or a ``Distribution`` that will be used to initialize the kernel.

    Notes
    -----
    As is typical in neural networks, this is technically correlation rather
    than convolution (because the kernel is not flipped).
    """

    output_shape = ChannelShapeParam("output_shape", low=1)

    _param_init_order = ["channels_last", "input_shape"]

    def __init__(
        self,
        n_filters,
        input_shape,
        output_shape=None,
        kernel_size=(3, 3),
        strides=(1, 1),
        padding="valid",
        channels_last=True,
        init=Uniform(-1, 1),
    ):
        super().__init__(
            n_filters=n_filters,
            input_shape=input_shape,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            channels_last=channels_last,
            init=init,
        )

        self.output_shape = (
            self._reverse_shape(self.input_shape.spatial_shape, self.n_filters)
            if output_shape is None
            else output_shape
        )

        if self.output_shape.dimensions != self.input_shape.dimensions:
            raise ValidationError(
                f"The number of dimensions ({self.output_shape.dimensions}) in the "
                f"provided `output_shape` {self.output_shape} does not match the number"
                f" of dimensions ({self.input_shape.dimensions}) in the input shape.",
                attr="output_shape",
                obj=self,
            )
        if self.output_shape.n_channels != self.n_filters:
            raise ValidationError(
                f"The number of channels in the provided `output_shape` "
                f"{self.output_shape} does not match the requested number "
                f"of filters ({self.n_filters}).",
                attr="output_shape",
                obj=self,
            )

        expected_input_shape = self._forward_shape(
            self.output_shape.spatial_shape, self.input_shape.n_channels
        )
        if self.input_shape != expected_input_shape:
            raise ValidationError(
                f"The provided `output_shape` {self.output_shape} would not produce "
                f"`input_shape` {self.input_shape} in a forward Convolution, "
                f"and is therefore not a valid output shape.",
                attr="output_shape",
                obj=self,
            )

    @property
    def _argreprs(self):
        argreprs = super()._argreprs
        argreprs.insert(2, f"output_shape={self.output_shape.shape}")
        return argreprs

    def _reverse_shape(self, input_spatial_shape, n_filters):
        output_shape = np.array(input_spatial_shape, dtype=rc.int_dtype)
        output_shape = 1 + (output_shape - 1) * self.strides
        if self.padding == "valid":
            output_shape += self.kernel_size
            output_shape -= 1

        return ChannelShape.from_space_and_channels(
            output_shape, n_filters, channels_last=self.channels_last
        )


class ChannelShape:
    """Represents shape information with variable channel position.

    .. versionadded:: 3.0.0

    Parameters
    ----------
    shape : iterable of int
        Signal shape
    channels_last : bool, optional
        If True (default), the last item in ``shape`` represents the channels,
        and the rest are spatial dimensions. Otherwise, the first item in
        ``shape`` is the channel dimension.
    """

    @classmethod
    def from_space_and_channels(cls, spatial_shape, n_channels, channels_last=True):
        """Create a ChannelShape from a spatial shape and number of channels.

        .. versionadded:: 3.2.0

        Parameters
        ----------
        spatial_shape : iterable of int
            The spatial part of the shape (not including channels).
        n_channels : int
            The number of channels.
        channels_last : bool, optional
            If True (default), the last item in ``shape`` represents the channels,
            and the rest are spatial dimensions. Otherwise, the first item in
            ``shape`` is the channel dimension.
        """
        shape = (
            tuple(spatial_shape) + (n_channels,)
            if channels_last
            else (n_channels,) + tuple(spatial_shape)
        )
        return cls(shape, channels_last=channels_last)

    def __init__(self, shape, channels_last=True):
        self.shape = tuple(shape)
        self.channels_last = channels_last

    def __eq__(self, other):
        return (
            type(self) == type(other)
            and self.shape == other.shape
            and self.channels_last == other.channels_last
        )

    def __hash__(self):
        return hash((self.shape, self.channels_last))

    def __repr__(self):
        return (
            f"{type(self).__name__}(shape={self.shape}, "
            f"channels_last={self.channels_last})"
        )

    def __str__(self):
        """Tuple-like string with channel position marked with 'ch'."""
        spatial = [str(s) for s in self.spatial_shape]
        channel = [f"ch={self.n_channels}"]
        return (
            "("
            + ", ".join(spatial + channel if self.channels_last else channel + spatial)
            + ")"
        )

    @property
    def spatial_shape(self):
        """The spatial part of the shape (omitting channels)."""
        if self.channels_last:
            return self.shape[:-1]
        return self.shape[1:]

    @property
    def size(self):
        """The total number of elements in the represented signal."""
        return np.prod(self.shape)

    @property
    def n_channels(self):
        """The number of channels in the represented signal."""
        return self.shape[-1 if self.channels_last else 0]

    @property
    def dimensions(self):
        """The spatial dimensionality of the represented signal."""
        return len(self.shape) - 1


class NoTransform(Transform):
    """Directly pass the signal through without any transform operations.

    .. versionadded:: 3.1.0

    Parameters
    ----------
    size_in : int
        Dimensionality of transform input and output.
    """

    def __init__(self, size_in):
        super().__init__()

        self._size_in = size_in

    def sample(self, rng=np.random):
        """Returns concrete weights to implement the specified transform.

        Parameters
        ----------
        rng : `numpy.random.RandomState`, optional
            Random number generator state.

        Raises
        ------
        TypeError
            There is nothing to sample for NoTransform, so it is an error
            if this is called.
        """
        raise TypeError("Cannot sample a NoTransform")

    @property
    def size_in(self):
        """Expected size of input to transform."""
        return self._size_in

    @property
    def size_out(self):
        """Expected size of output from transform."""
        return self._size_in


# aliases
Conv = Convolution
ConvTranspose = ConvolutionTranspose



import sys
import weakref

import numpy as np
import pytest

from nengo.utils.stdlib import (
    FrozenOrderedSet,
    OrderedSet,
    Timer,
    WeakKeyDefaultDict,
    WeakKeyIDDictionary,
    WeakSet,
    checked_call,
    groupby,
)


def test_checked_call():
    def func1(a):
        return a

    def func2(a, b=0, **kwargs):
        return a + b

    def func3(
        a, b=0, c=0, *args, **kwargs
    ):  # pylint: disable=keyword-arg-before-vararg
        return a + b + c + sum(args)

    func4 = lambda x=[0]: sum(x)

    class A:
        def __call__(self, a, b):
            return a + b

    assert checked_call(func1) == (None, False)
    assert checked_call(func1, 1) == (1, True)
    assert checked_call(func1, 1, 2) == (None, False)
    assert checked_call(func1, 1, two=2) == (None, False)

    assert checked_call(func2) == (None, False)
    assert checked_call(func2, 1) == (1, True)
    assert checked_call(func2, 1, 2) == (3, True)
    assert checked_call(func2, 1, 2, three=3) == (3, True)
    assert checked_call(func2, 1, 2, 3) == (None, False)

    assert checked_call(func3) == (None, False)
    assert checked_call(func3, 1) == (1, True)
    assert checked_call(func3, 1, 2) == (3, True)
    assert checked_call(func3, 1, 2, 3) == (6, True)
    assert checked_call(func3, 1, 2, 3, 4, 5, 6) == (21, True)

    assert checked_call(func4) == (0, True)
    assert checked_call(func4, [1, 2]) == (3, True)
    assert checked_call(func4, [1], 2) == (None, False)

    assert checked_call(A(), 1) == (None, False)
    assert checked_call(A(), 1, 2) == (3, True)
    assert checked_call(A(), 1, 2, 3) == (None, False)

    assert checked_call(np.sin) == (None, False)
    assert checked_call(np.sin, 0) == (0, True)
    assert checked_call(np.sin, 0, np.array([1.0])) == (np.array([0.0]), True)
    assert checked_call(np.sin, 0, np.array([1.0]), 1) == (None, False)


def test_checked_call_errors():
    class A:
        def __call__(self, a):
            raise NotImplementedError()

    assert checked_call(A()) == (None, False)
    with pytest.raises(NotImplementedError):
        checked_call(A(), 1)

    assert checked_call(np.sin, 1, 2, 3) == (None, False)
    with pytest.raises((ValueError, TypeError)):
        checked_call(lambda x: np.sin(1, 2, 3), 1)


@pytest.mark.parametrize(
    "hashable, force_list", [(False, False), (False, True), (True, False), (True, True)]
)
def test_groupby(hashable, force_list, rng):
    if hashable:
        keys = list(range(1, 5))
    else:
        keys = [[0, 0], [0, 1], [1, 0], [1, 1]]

    keys = sorted(keys)

    # make groups and pairs
    groups = [rng.randn(rng.randint(5, 10)) for _ in keys]

    pairs = []
    for key, group in zip(keys, groups):
        pairs.extend((key, value) for value in group)

    # shuffle pairs
    pairs = [pairs[i] for i in rng.permutation(len(pairs))]

    # call groupby
    keygroups = groupby(pairs, lambda p: p[0], force_list=force_list)

    keys2 = sorted([x[0] for x in keygroups])
    assert keys2 == keys

    for key2, keygroup2 in keygroups:
        group = groups[keys.index(key2)]
        group2 = [x[1] for x in keygroup2]
        assert sorted(group2) == sorted(group)


def test_timer():
    with Timer() as timer:
        for _ in range(1000):
            _ = 2 + 2
    assert timer.duration > 0.0
    assert timer.duration < 1.0  # Pretty bad worst case


class C:
    def method(self):
        pass


def test_weakkeydefaultdict():
    factory = lambda: "default"
    d = WeakKeyDefaultDict(factory)
    o = C()

    assert len(d) == 0
    assert d[o] == "default"

    d[o] = "changed"
    assert len(d) == 1
    assert d[o] == "changed"

    del d[o]
    assert len(d) == 0
    assert o not in d

    d[o] = "changed"
    for index in d:
        assert index == o
    del index  # pylint: disable=undefined-loop-variable
    del o
    assert len(d) == 0


def test_weakkeyiddictionary_init():
    obj = C()
    val = 364

    # construct from dictionary
    d = WeakKeyIDDictionary({obj: val})
    assert d[obj] == val

    # construct from another WeakKeyIDDictionary
    d2 = WeakKeyIDDictionary(d)
    assert d2[obj] == val


def test_weakkeydict_popitem(key1=C(), key2=C(), val1="v1", val2="v2"):
    d = WeakKeyIDDictionary()
    d[key1] = val1
    d[key2] = val2

    assert len(d) == 2
    k1, v1 = d.popitem()
    assert len(d) == 1
    assert k1 in (key1, key2) and v1 is (val1 if k1 is key1 else val2)

    k2, v2 = d.popitem()
    assert len(d) == 0
    assert k2 is (key2 if k1 is key1 else key1) and v2 is (val1 if k2 is key1 else val2)


def test_weakkeydict_setdefault(key=C(), value1="v1", value2="v2"):
    d = WeakKeyIDDictionary()
    o = d.setdefault(key, value1)
    assert o is value1
    assert key in d
    assert d.get(key) is value1
    assert d[key] is value1

    o = d.setdefault(key, value2)
    assert o is value1
    assert key in d
    assert d.get(key) is value1
    assert d[key] is value1


def test_weakkeydict_update():
    """This exercises d.update(), len(d), d.keys(), in d,  d.get(), d[]."""
    in_d = {C(): 1, C(): 2, C(): 3}
    d = WeakKeyIDDictionary()
    d.update(in_d)

    assert len(d) == len(in_d)
    for k in d.keys():  # pylint: disable=consider-using-dict-items
        assert k in in_d, "mysterious new key appeared in weak dict"
        v = in_d.get(k)
        assert v is d[k] and v is d.get(k)
    for k in in_d:  # pylint: disable=consider-using-dict-items
        assert k in d, "original key disappeared in weak dict"
        v = in_d[k]
        assert v is d[k] and v is d.get(k)


def test_weakkeyiddict_contains_none():
    assert (None in WeakKeyIDDictionary()) is False


def test_weakkeydict_delitem():
    d = WeakKeyIDDictionary()
    o1 = C()
    o2 = C()
    d[o1] = "something"
    d[o2] = "something"
    assert len(d) == 2
    del d[o1]
    assert len(d) == 1
    assert list(d.keys()) == [o2]


def test_weakkeydict_bad_delitem():
    d = WeakKeyIDDictionary()
    o = C()

    # An attempt to delete an object that isn't there should raise KeyError.
    with pytest.raises(KeyError):
        del d[o]
    with pytest.raises(KeyError):
        print(d[o])

    # If a key isn't of a weakly referencable type, __getitem__ and __delitem__ raise
    # keyerror, __setitem__ raises a TypeError
    with pytest.raises(KeyError):
        del d[13]
    with pytest.raises(KeyError):
        print(d[13])
    with pytest.raises(TypeError):
        d[13] = 13


def test_weakkeydict_frees_values():
    d = WeakKeyIDDictionary()
    k = C()
    v = C()
    d[k] = v
    weak_v = weakref.ref(v)
    del v
    assert sys.getrefcount(weak_v()) > 1  # function argument might make it > 1
    del k
    v = weak_v()
    assert v is None, "Value in WeakKeyIDDictionary not garbage collected."


def test_weakset():
    s = WeakSet()
    k = C()

    s.add(k)
    assert len(s) == 1
    assert k in s

    s.discard(k)
    assert len(s) == 0
    assert k not in s

    s.add(k)
    del k
    assert len(s) == 0


def test_weakset_init():
    k = C()

    # init from a list
    s = WeakSet([k])
    assert k in s

    # init from another weakset
    s2 = WeakSet(s)
    assert k in s2


@pytest.mark.skipif(
    sys.version_info < (3, 6, 0), reason="OrderedSet only works in Python>=3.6"
)
def test_ordered_set():
    my_set = OrderedSet(("c", "b", "a"))

    # order is preserved
    assert tuple(my_set) == ("c", "b", "a")
    my_set |= "d"
    assert tuple(my_set) == ("c", "b", "a", "d")
    my_set -= "b"
    assert tuple(my_set) == ("c", "a", "d")

    # not hashable
    with pytest.raises(TypeError, match="OrderedSet is not hashable"):
        _ = {my_set: "val"}

    # hashable
    my_frozen_set = FrozenOrderedSet(my_set)
    assert tuple(my_frozen_set) == ("c", "a", "d")
    assert {my_frozen_set: "val"}[my_frozen_set] == "val"



import pickle

import numpy as np
import pytest

import nengo
from nengo.builder import Model
from nengo.builder.signal import Signal, SignalDict
from nengo.exceptions import SignalError
from nengo.utils.numpy import scipy_sparse


def test_signaldict(allclose):
    """Tests SignalDict's dict overrides."""
    signaldict = SignalDict()

    scalar = Signal(1.0)

    # Both __getitem__ and __setitem__ raise KeyError
    with pytest.raises(KeyError):
        print(signaldict[scalar])
    with pytest.raises(KeyError):
        signaldict[scalar] = np.array(1.0)

    signaldict.init(scalar)

    # tests repeat init
    with pytest.raises(SignalError, match="Cannot add signal twice"):
        signaldict.init(scalar)

    assert allclose(signaldict[scalar], np.array(1.0))
    # __getitem__ handles scalars
    assert signaldict[scalar].shape == ()

    one_d = Signal([1.0])
    signaldict.init(one_d)
    assert allclose(signaldict[one_d], np.array([1.0]))
    assert signaldict[one_d].shape == (1,)

    two_d = Signal([[1.0], [1.0]])
    signaldict.init(two_d)
    assert allclose(signaldict[two_d], np.array([[1.0], [1.0]]))
    assert signaldict[two_d].shape == (2, 1)

    # __getitem__ handles views implicitly (note no .init)
    two_d_view = two_d[0, :]
    assert allclose(signaldict[two_d_view], np.array([1.0]))
    assert signaldict[two_d_view].shape == (1,)

    # __setitem__ ensures memory location stays the same
    memloc = signaldict[scalar].__array_interface__["data"][0]
    signaldict[scalar] = np.array(0.0)
    assert allclose(signaldict[scalar], np.array(0.0))
    assert signaldict[scalar].__array_interface__["data"][0] == memloc

    memloc = signaldict[one_d].__array_interface__["data"][0]
    signaldict[one_d] = np.array([0.0])
    assert allclose(signaldict[one_d], np.array([0.0]))
    assert signaldict[one_d].__array_interface__["data"][0] == memloc

    memloc = signaldict[two_d].__array_interface__["data"][0]
    signaldict[two_d] = np.array([[0.0], [0.0]])
    assert allclose(signaldict[two_d], np.array([[0.0], [0.0]]))
    assert signaldict[two_d].__array_interface__["data"][0] == memloc

    # __str__ pretty-prints signals and current values
    # Order not guaranteed for dicts, so we have to loop
    for k in signaldict:
        assert f"{repr(k)} {repr(signaldict[k])}" in str(signaldict)


def test_signaldict_reset(allclose):
    """Tests SignalDict's reset function."""
    signaldict = SignalDict()
    two_d = Signal([[1.0], [1.0]])
    signaldict.init(two_d)

    two_d_view = two_d[0, :]
    signaldict.init(two_d_view)

    signaldict[two_d_view] = -0.5
    assert allclose(signaldict[two_d], np.array([[-0.5], [1]]))

    signaldict[two_d] = np.array([[-1], [-1]])
    assert allclose(signaldict[two_d], np.array([[-1], [-1]]))
    assert allclose(signaldict[two_d_view], np.array([-1]))

    signaldict.reset(two_d_view)
    assert allclose(signaldict[two_d_view], np.array([1]))
    assert allclose(signaldict[two_d], np.array([[1], [-1]]))

    signaldict.reset(two_d)
    assert allclose(signaldict[two_d], np.array([[1], [1]]))


def test_assert_named_signals():
    """Make sure assert_named_signals works."""
    Signal(np.array(0.0))
    Signal.assert_named_signals = True
    with pytest.raises(AssertionError):
        Signal(np.array(0.0))

    # So that other tests that build signals don't fail...
    Signal.assert_named_signals = False


def test_signal_values(allclose):
    """Make sure Signal.initial_value works."""
    two_d = Signal([[1.0], [1.0]])
    assert allclose(two_d.initial_value, np.array([[1], [1]]))
    two_d_view = two_d[0, :]
    assert allclose(two_d_view.initial_value, np.array([1]))

    # cannot change signal value after creation
    with pytest.raises(SignalError):
        two_d.initial_value = np.array([[0.5], [-0.5]])
    with pytest.raises((ValueError, RuntimeError)):
        two_d.initial_value[...] = np.array([[0.5], [-0.5]])


def test_signal_reshape():
    """Tests Signal.reshape"""
    # check proper shape after reshape
    three_d = Signal(np.ones((2, 2, 2)))
    assert three_d.reshape((8,)).shape == (8,)
    assert three_d.reshape((4, 2)).shape == (4, 2)
    assert three_d.reshape((2, 4)).shape == (2, 4)
    assert three_d.reshape(-1).shape == (8,)
    assert three_d.reshape((4, -1)).shape == (4, 2)
    assert three_d.reshape((-1, 4)).shape == (2, 4)
    assert three_d.reshape((2, -1, 2)).shape == (2, 2, 2)
    assert three_d.reshape((1, 2, 1, 2, 2, 1)).shape == (1, 2, 1, 2, 2, 1)

    # check with non-contiguous arrays (and with offset)
    value = np.arange(20).reshape(5, 4)
    s = Signal(np.array(value), name="s")

    s0slice = slice(0, 3), slice(None, None, 2)
    s0shape = 2, 3
    s0 = s[s0slice].reshape(*s0shape)
    assert s.offset == 0
    assert np.array_equal(s0.initial_value, value[s0slice].reshape(*s0shape))

    s1slice = slice(1, None), slice(None, None, 2)
    s1shape = 2, 4
    s1 = s[s1slice].reshape(s1shape)
    assert s1.offset == 4 * s1.dtype.itemsize
    assert np.array_equal(s1.initial_value, value[s1slice].reshape(s1shape))

    # check error if non-contiguous array cannot be reshaped without copy
    s2slice = slice(None, None, 2), slice(None, None, 2)
    s2shape = 2, 3
    s2 = s[s2slice]
    with pytest.raises(SignalError):
        s2.reshape(s2shape)

    # check that views are working properly (incrementing `s` effects views)
    values = SignalDict()
    values.init(s)
    values.init(s0)
    values.init(s1)

    values[s] += 1
    assert np.array_equal(values[s0], value[s0slice].reshape(s0shape) + 1)
    assert np.array_equal(values[s1], value[s1slice].reshape(s1shape) + 1)


def test_signal_slicing(rng):
    slices = [
        0,
        1,
        slice(None, -1),
        slice(1, None),
        slice(1, -1),
        slice(None, None, 3),
        slice(1, -1, 2),
    ]

    x = np.arange(12, dtype=float)
    y = np.arange(24, dtype=float).reshape(4, 6)
    a = Signal(x.copy())
    b = Signal(y.copy())

    for _ in range(100):
        si0, si1 = rng.randint(0, len(slices), size=2)
        s0, s1 = slices[si0], slices[si1]
        assert np.array_equiv(a[s0].initial_value, x[s0])
        assert np.array_equiv(b[s0, s1].initial_value, y[s0, s1])

    with pytest.raises(ValueError):
        print(a[[0, 2]])
    with pytest.raises(ValueError):
        print(b[[0, 1], [3, 4]])


def test_commonsig_readonly():
    """Test that the common signals cannot be modified."""
    net = nengo.Network(label="test_commonsig")
    model = Model()
    model.build(net)
    signals = SignalDict()

    for sig in model.sig["common"].values():
        signals.init(sig)
        with pytest.raises((ValueError, RuntimeError)):
            signals[sig] = np.array([-1])
        with pytest.raises((ValueError, RuntimeError)):
            signals[sig][...] = np.array([-1])


def test_signal_offset():
    value = np.eye(3)
    s = Signal(value)
    assert s.offset == 0
    assert s[1].offset == value.strides[0]

    value_view = value[1]
    s = Signal(value_view)
    assert s.offset == 0
    assert s[0:].offset == 0
    assert s[1:].offset == value.strides[1]


def test_may_share_memory(rng):
    def may_share_memory(a, b):
        a_b, b_a = a.may_share_memory(b), b.may_share_memory(a)
        assert a_b == b_a  # check commutative property
        return a_b

    sig_a = Signal(initial_value=rng.rand(3, 4))
    sig_b = Signal(initial_value=rng.rand(3, 4))
    assert may_share_memory(sig_a, sig_a)
    assert may_share_memory(sig_b, sig_b)
    assert not may_share_memory(sig_a, sig_b)

    assert may_share_memory(sig_a[0:2, :], sig_a)
    assert may_share_memory(sig_a[0:2, :], sig_a[1:3, :])
    assert not may_share_memory(sig_a[0, :], sig_a[1, :])
    assert not may_share_memory(sig_a[0:2, :], sig_b)
    assert not may_share_memory(sig_a[0:2, :], sig_b[1:3, :])


def make_signal(sig_type, shape, indices, data):
    dense = np.zeros(shape)
    dense[indices[:, 0], indices[:, 1]] = data
    if sig_type == "sparse_scipy":
        m = scipy_sparse.csr_matrix((data, indices.T), shape=shape)
    elif sig_type == "sparse_nengo":
        m = nengo.transforms.SparseMatrix(data=data, indices=indices, shape=shape)
    else:
        m = dense
    return Signal(m, name="MySignal"), dense


@pytest.mark.parametrize("sig_type", ("dense", "sparse_scipy", "sparse_nengo"))
def test_signal_initial_value(sig_type, tmp_path, allclose):
    if sig_type == "sparse_scipy":
        pytest.importorskip("scipy.sparse")

    sig, dense = make_signal(
        sig_type,
        shape=(3, 3),
        indices=np.asarray([[0, 0], [0, 2], [1, 1], [2, 2]]),
        data=[1.0, 2.0, 1.0, 1.5],
    )

    # check initial_value equality
    assert allclose(
        sig.initial_value.toarray()
        if sig_type.startswith("sparse")
        else sig.initial_value,
        dense,
    )

    # cannot change once set
    with pytest.raises(SignalError, match="Cannot change initial value"):
        sig.initial_value = sig.initial_value

    # check signal pickles correctly
    pkl_path = tmp_path / "tmp.pkl"
    with open(pkl_path, "wb") as f:
        pickle.dump(sig, f)

    with open(pkl_path, "rb") as f:
        pkl_sig = pickle.load(f)

    # initial_value still matches after pickle/unpickle
    assert allclose(
        sig.initial_value.toarray()
        if sig_type.startswith("sparse")
        else sig.initial_value,
        pkl_sig.initial_value.toarray()
        if sig_type.startswith("sparse")
        else pkl_sig.initial_value,
    )


@pytest.mark.parametrize("sig_type", ("dense", "sparse_scipy", "sparse_nengo"))
def test_signal_slice_reshape(sig_type):
    if sig_type == "sparse_scipy":
        pytest.importorskip("scipy.sparse")

    sig, _ = make_signal(
        sig_type,
        shape=(3, 3),
        indices=np.asarray([[0, 0], [0, 2], [1, 1], [2, 2]]),
        data=[1.0, 2.0, 1.0, 1.5],
    )

    # check slicing
    if sig_type == "dense":
        sig_slice = sig[:2]
        assert sig_slice.shape == (2, 3)
        assert sig_slice.base is sig
        assert sig.may_share_memory(sig_slice)
    else:
        with pytest.raises(SignalError, match="sparse Signal"):
            print(sig[:2])

    # check reshaping
    if sig_type == "dense":
        sig_reshape = sig.reshape((1, 9))
        assert sig_reshape.shape == (1, 9)
        assert sig_reshape.base is sig
        assert sig.may_share_memory(sig_reshape)
    else:
        with pytest.raises(SignalError, match="sparse Signal"):
            sig.reshape((1, 9))


@pytest.mark.parametrize("sig_type", ("dense", "sparse_scipy", "sparse_nengo"))
def test_signal_properties(sig_type):
    if sig_type == "sparse_scipy":
        pytest.importorskip("scipy.sparse")

    sig, _ = make_signal(
        sig_type,
        shape=(3, 3),
        indices=np.asarray([[0, 0], [0, 2], [1, 1], [2, 2]]),
        data=[1.0, 2.0, 1.0, 1.5],
    )

    # check properties
    assert sig.base is sig
    assert sig.dtype == sig.initial_value.dtype == np.float64
    assert sig.elemoffset == 0
    assert sig.elemstrides == ((3, 1) if sig_type == "dense" else None)
    assert not sig.is_view
    assert sig.itemsize == 8
    assert sig.name == "MySignal"
    assert sig.nbytes == sig.itemsize * sig.size
    assert sig.ndim == 2
    assert sig.offset == 0
    assert not sig.readonly
    assert sig.shape == (3, 3)
    assert sig.size == (
        np.sum(sig.initial_value.toarray() != 0)
        if sig_type.startswith("sparse")
        else np.prod((3, 3))
    )
    assert sig.strides == (
        (3 * sig.itemsize, sig.itemsize) if sig_type == "dense" else None
    )
    assert sig.sparse if sig_type.startswith("sparse") else not sig.sparse

    # modifying properties
    sig.name = "NewName"
    assert sig.name == "NewName"
    sig.readonly = True
    assert sig.readonly


@pytest.mark.parametrize("sig_type", ("dense", "sparse_scipy", "sparse_nengo"))
def test_signal_init(sig_type):
    if sig_type == "sparse_scipy":
        pytest.importorskip("scipy.sparse")

    sig, dense = make_signal(
        sig_type,
        shape=(3, 3),
        indices=np.asarray([[0, 0], [0, 2], [1, 1], [2, 2]]),
        data=[1.0, 2.0, 1.0, 1.5],
    )
    signals = SignalDict()
    signals.init(sig)
    assert np.all(signals[sig] == dense)

    sig.readonly = True
    signals = SignalDict()
    signals.init(sig)
    with pytest.raises((ValueError, RuntimeError, TypeError)):
        signals[sig].data[0] = -1


def test_signal_shape():
    shape = (3, 4)
    sig = Signal(shape=shape)
    assert sig.shape == shape

    Signal(np.zeros(shape), shape=shape)
    with pytest.raises(AssertionError):
        Signal(np.zeros((2, 3)), shape=shape)


def tests_signal_nan():
    with_nan = np.ones(4)
    with_nan[1] = np.nan
    with pytest.raises(SignalError, match="contains NaNs"):
        Signal(initial_value=with_nan)



from nengo.connection import Connection
from nengo.ensemble import Ensemble
from nengo.exceptions import ObsoleteError
from nengo.network import Network
from nengo.node import Node


class Oscillator(Network):
    """A two-dimensional ensemble with interacting recurrent connections.

    The ensemble connects to itself in a manner similar to the integrator;
    however, here the two dimensions interact with each other to implement
    a cyclic oscillator.

    Parameters
    ----------
    recurrent_tau : float
        Time constant on the recurrent connection.
    frequency : float
        Desired frequency, in radians per second, of the cyclic oscillation.
    n_neurons : int
        Number of neurons in the recurrently connected ensemble.
    **kwargs
        Keyword arguments passed through to ``nengo.Network``
        like 'label' and 'seed'.

    Attributes
    ----------
    ensemble : Ensemble
        The recurrently connected oscillatory ensemble.
    input : Node
        Provides the input signal.
    """

    def __init__(self, recurrent_tau, frequency, n_neurons, **kwargs):
        if "net" in kwargs:
            raise ObsoleteError("The 'net' argument is no longer supported.")
        kwargs.setdefault("label", "Oscillator")
        super().__init__(**kwargs)

        with self:
            self.input = Node(label="In", size_in=2)
            self.ensemble = Ensemble(n_neurons, dimensions=2, label="Oscillator")

            tA = [[1, -frequency * recurrent_tau], [frequency * recurrent_tau, 1]]
            Connection(
                self.ensemble, self.ensemble, synapse=recurrent_tau, transform=tA
            )
            Connection(self.input, self.ensemble, synapse=None)

        self.output = self.ensemble



import numpy as np
import pytest

from nengo._vendor.npconv2d import conv2d
from nengo.builder.signal import Signal
from nengo.builder.tests.test_operator import _test_operator_arg_attributes
from nengo.builder.transforms import ConvInc, ConvTransposeInc
from nengo.transforms import ChannelShape, Convolution, ConvolutionTranspose


@pytest.mark.parametrize("channels_last", (True, False))
@pytest.mark.parametrize("stride0", (1, 2))
@pytest.mark.parametrize("stride1", (1, 2))
@pytest.mark.parametrize("kernel0", (4, 5))
@pytest.mark.parametrize("kernel1", (4, 5))
@pytest.mark.parametrize("padding", ("same", "valid"))
def test_convinc_2d(
    channels_last, stride0, stride1, kernel0, kernel1, padding, rng, allclose
):
    correlate2d = pytest.importorskip("scipy.signal").correlate2d

    shape0 = 16
    shape1 = 17
    in_channels = 32
    out_channels = 64
    x_shape = (
        (shape0, shape1, in_channels)
        if channels_last
        else (in_channels, shape0, shape1)
    )
    x = Signal(rng.randn(*x_shape))
    w = Signal(rng.randn(kernel0, kernel1, in_channels, out_channels))

    conv = Convolution(
        out_channels,
        x_shape,
        kernel_size=(kernel0, kernel1),
        strides=(stride0, stride1),
        padding=padding,
        channels_last=channels_last,
    )

    y = Signal(np.zeros(conv.output_shape.shape))

    signals = {sig: np.array(sig.initial_value) for sig in (x, w, y)}
    step = ConvInc(w, x, y, conv).make_step(signals, None, None)

    step()

    x0 = x.initial_value

    if not channels_last:
        x0 = np.moveaxis(x0, 0, -1)

    if padding == "same":
        strides = np.asarray([stride0, stride1])
        padding = np.ceil(np.asarray([shape0, shape1]) / strides)
        padding = np.maximum(
            (padding - 1) * strides + (kernel0, kernel1) - (shape0, shape1), 0
        ).astype(np.int64)
        x0 = np.pad(
            x0,
            [
                (padding[0] // 2, padding[0] - padding[0] // 2),
                (padding[1] // 2, padding[1] - padding[1] // 2),
                (0, 0),
            ],
            "constant",
        )

    y0 = np.stack(
        [
            np.sum(
                [
                    correlate2d(x0[..., j], w.initial_value[..., j, i], mode="valid")
                    for j in range(in_channels)
                ],
                axis=0,
            )
            for i in range(out_channels)
        ],
        axis=-1,
    )
    y0 = y0[::stride0, ::stride1, :]
    if not channels_last:
        y0 = np.moveaxis(y0, -1, 0)

    assert allclose(signals[y], y0)


@pytest.mark.parametrize("channels_last", (True, False))
@pytest.mark.parametrize("strides", [(1, 1), (2, 2), (3, 2)])
@pytest.mark.parametrize("kernel_size", [(4, 5), (5, 4), (5, 5)])
@pytest.mark.parametrize("padding", ("same", "valid"))
def test_convtransposeinc_2d(
    channels_last, strides, kernel_size, padding, rng, allclose, plt
):
    """Test ConvTransposeInc by ensuring it is the transpose of ConvInc.

    Since convolution is a linear operator, it can be expressed as a matrix ``A``.
    We can therefore state that ``C.dot(A.dot(x)) == (A.T.dot(C.T)).T.dot(x)``,
    for an arbitrary vector ``x`` and arbitrary matrix ``C``.

    This test asserts this identity, and thereby tests the ``ConvTransposeInc`` operator
    against the ``ConvInc`` operator.
    """
    spatial_shape = (16, 17)
    in_channels = 32
    out_channels = 64

    x_shape = ChannelShape.from_space_and_channels(
        spatial_shape, in_channels, channels_last=channels_last
    )
    conv = Convolution(
        out_channels,
        x_shape,
        kernel_size=kernel_size,
        strides=strides,
        padding=padding,
        channels_last=channels_last,
    )

    nk = 10  # number of vectors we test ConvTransposeInc with
    C = rng.randn(nk, conv.output_shape.size)

    # compute ``conv_output = C.dot(A.dot(x))``, where ``A`` is the convolution operator
    x = Signal(rng.randn(*x_shape.shape))
    w = Signal(rng.randn(kernel_size[0], kernel_size[1], in_channels, out_channels))
    y = Signal(np.zeros(conv.output_shape.shape))

    signals = {sig: np.array(sig.initial_value) for sig in (x, w, y)}
    step_conv = ConvInc(w, x, y, conv).make_step(signals, None, None)
    step_conv()

    x_flat = signals[x].ravel()
    y_flat = signals[y].ravel()
    conv_output = C.dot(y_flat)

    # compute ``conv_transpose_output = (A.T.dot(C.T)).T.dot(x)``, where ``A.T`` is the
    # transpose convolution operator (applied to one column of ``C.T`` at a time)
    conv_transpose = ConvolutionTranspose(
        in_channels,
        conv.output_shape,
        output_shape=x_shape,
        kernel_size=kernel_size,
        strides=strides,
        padding=padding,
        channels_last=channels_last,
    )
    xt = Signal(np.zeros(conv_transpose.input_shape.shape))
    wt = Signal(np.transpose(w.initial_value, (0, 1, 3, 2)))
    yt = Signal(np.zeros(conv_transpose.output_shape.shape))

    signals_tr = {sig: np.array(sig.initial_value) for sig in (xt, wt, yt)}
    step_trans = ConvTransposeInc(wt, xt, yt, conv_transpose).make_step(
        signals_tr, None, None
    )

    AtCt = []
    for k in range(nk):
        signals_tr[xt][:] = C[k].reshape(conv_transpose.input_shape.shape)
        signals_tr[yt][:] = 0
        step_trans()
        AtCt.append(signals_tr[yt].copy().ravel())

    AtCt = np.array(AtCt)
    conv_transpose_output = AtCt.dot(x_flat)
    assert conv_transpose_output.shape == conv_output.shape

    success = allclose(conv_transpose_output, conv_output)
    if success:
        plt.saveas = None
    else:
        debug_convtransposeinc_2d(w, wt, x, xt, y, yt, conv, conv_transpose, plt)
    assert success


def debug_convtransposeinc_2d(w, wt, x, xt, y, yt, conv, conv_transpose, plt):
    in_channels = w.shape[2]
    channels_last = conv.channels_last

    # transpose convolution should act as the derivative for forward convolution,
    # compute and compare to the derivative to see exactly where the difference is
    signals = {sig: np.array(sig.initial_value) for sig in (x, w, y)}
    step_conv = ConvInc(w, x, y, conv).make_step(signals, None, None)
    step_conv()

    dy_dx = []
    for j in range(conv.input_shape.size):
        xj = np.zeros(conv.input_shape.shape)
        xj.ravel()[j] = 1

        signals[x][:] = xj
        signals[y][:] = 0
        step_conv()

        dy_dx.append(signals[y].copy().ravel())

    dy_dx = np.column_stack(dy_dx)
    assert dy_dx.shape == (conv.output_shape.size, conv.input_shape.size)

    signals_tr = {sig: np.array(sig.initial_value) for sig in (xt, wt, yt)}
    step_trans = ConvTransposeInc(wt, xt, yt, conv_transpose).make_step(
        signals_tr, None, None
    )

    for i in range(conv.output_shape.size):
        xi = np.zeros(conv.output_shape.shape)
        xi.ravel()[i] = 1

        signals_tr[xt][:] = xi
        signals_tr[yt][:] = 0
        step_trans()

        actual = signals_tr[yt]
        expected = dy_dx[i, :].reshape(conv.input_shape.shape)

        assert actual.shape == expected.shape
        if not np.allclose(actual, expected):
            for k in range(in_channels):
                actual_k = actual[:, :, k] if channels_last else actual[k]
                expected_k = expected[:, :, k] if channels_last else expected[k]

                if not np.allclose(actual, expected):
                    plt.subplot(121)
                    plt.imshow(expected_k, vmin=expected_k.min(), vmax=expected_k.max())
                    plt.title(f"expected out={i} ch={k}")

                    plt.subplot(122)
                    plt.imshow(actual_k, vmin=expected_k.min(), vmax=expected_k.max())
                    plt.title(f"actual out={i} ch={k}")

                    assert False, f"Output {i} channel {k} differs"


@pytest.mark.parametrize("x_size", (1, 2, 3, 4))
@pytest.mark.parametrize("k_size", (1, 2, 3, 4))
@pytest.mark.parametrize("stride", (1, 2, 3, 4))
@pytest.mark.parametrize("padding", ("same", "valid"))
def test_conv2d_gradx(padding, stride, k_size, x_size, rng, plt, allclose):
    tf = pytest.importorskip("tensorflow")

    in_channels = 1
    out_channels = 1

    if padding == "same":
        y_min = (x_size - 1) * stride + 1
        y_max = x_size * stride
    else:
        y_min = (x_size - 1) * stride + k_size
        y_max = x_size * stride + k_size - 1

    bad_sizes = []
    for y_size in range(y_min, y_max + 1):
        x_shape = (1, x_size, x_size, in_channels)
        y_shape = (1, y_size, y_size, out_channels)
        k_shape = (k_size, k_size, out_channels, in_channels)

        x = rng.uniform(-1, 1, size=x_shape)
        kernel = rng.uniform(-1, 1, size=k_shape)
        y_tf = tf.nn.conv2d_transpose(
            x, kernel, y_shape, stride, padding=padding.upper()
        ).numpy()[0]

        kernel_t = np.transpose(kernel, (0, 1, 3, 2))
        y_np = conv2d.conv2d_gradx(
            kernel_t,
            x,
            xsize=y_shape[1:3],
            pad=padding.upper(),
            stride=(stride, stride),
        )[0]

        assert y_np.shape == y_tf.shape

        if not allclose(y_np, y_tf):
            plt.subplot(1, 2, 1)
            plt.imshow(y_tf[..., 0], vmin=y_tf.min(), vmax=y_tf.max())
            plt.subplot(1, 2, 2)
            plt.imshow(y_np[..., 0], vmin=y_tf.min(), vmax=y_tf.max())
            bad_sizes.append(y_size)

    if len(bad_sizes) == 0:
        plt.saveas = None
    assert len(bad_sizes) == 0, (y_min, y_max, bad_sizes)


def test_convinc_attrs_decstr():
    argnames = ["W", "X", "Y", "conv"]
    non_signals = ["conv"]

    conv = Convolution(4, (3, 5, 2))
    _, sim = _test_operator_arg_attributes(
        ConvInc, argnames, non_signals=non_signals, args={"conv": conv}
    )
    assert str(sim) == "ConvInc{conv2d(W, X) -> Y}"

    conv_transpose = ConvolutionTranspose(4, (3, 5, 2))
    _, sim = _test_operator_arg_attributes(
        ConvTransposeInc,
        argnames,
        non_signals=non_signals,
        args={"conv": conv_transpose},
    )
    assert str(sim) == "ConvTransposeInc{convtranspose2d(W, X) -> Y}"



import re

import numpy as np
import pytest

from nengo.exceptions import SpaParseError, ValidationError
from nengo.spa import Vocabulary


def test_add(rng, allclose):
    v = Vocabulary(3, rng=rng)
    v.add("A", [1, 2, 3])
    v.add("B", [4, 5, 6])
    v.add("C", [7, 8, 9])
    assert allclose(v.vectors, [[1, 2, 3], [4, 5, 6], [7, 8, 9]])


def test_include_pairs(rng):
    v = Vocabulary(10, rng=rng)
    v["A"]
    v["B"]
    v["C"]
    assert v.key_pairs is None
    v.include_pairs = True
    assert v.key_pairs == ["A*B", "A*C", "B*C"]
    v.include_pairs = False
    assert v.key_pairs is None
    v.include_pairs = True
    v["D"]
    assert v.key_pairs == ["A*B", "A*C", "B*C", "A*D", "B*D", "C*D"]

    v = Vocabulary(12, include_pairs=True, rng=rng)
    v["A"]
    v["B"]
    v["C"]
    assert v.key_pairs == ["A*B", "A*C", "B*C"]


def test_parse(rng, allclose):
    v = Vocabulary(64, rng=rng)
    A = v.parse("A")
    B = v.parse("B")
    C = v.parse("C")
    assert allclose((A * B).v, v.parse("A * B").v)
    assert allclose((A * ~B).v, v.parse("A * ~B").v)
    assert allclose((A + B).v, v.parse("A + B").v)
    assert allclose((A - (B * C) * 3 + ~C).v, v.parse("A-(B*C)*3+~C").v)

    assert allclose(v.parse("0").v, np.zeros(64))
    assert allclose(v.parse("1").v, np.eye(64)[0])
    assert allclose(v.parse("1.7").v, np.eye(64)[0] * 1.7)

    with pytest.raises(SyntaxError):
        v.parse("A((")
    with pytest.raises(SpaParseError):
        v.parse('"hello"')


def test_invalid_dimensions():
    with pytest.raises(ValidationError):
        Vocabulary(1.5)
    with pytest.raises(ValidationError):
        Vocabulary(0)
    with pytest.raises(ValidationError):
        Vocabulary(-1)


def test_identity(rng, allclose):
    v = Vocabulary(64, rng=rng)
    assert allclose(v.identity.v, np.eye(64)[0])


def test_text(rng):
    v = Vocabulary(64, rng=rng)
    x = v.parse("A+B+C")
    y = v.parse("-D-E-F")
    ptr = r"-?[01]\.[0-9]{2}[A-F]"
    assert re.match(";".join([ptr] * 3), v.text(x))
    assert re.match(";".join([ptr] * 2), v.text(x, maximum_count=2))
    assert re.match(ptr, v.text(x, maximum_count=1))
    assert len(v.text(x, maximum_count=10).split(";")) <= 10
    assert re.match(";".join([ptr] * 4), v.text(x, minimum_count=4))
    assert re.match(
        ";".join([ptr.replace("F", "C")] * 3),
        v.text(x, minimum_count=4, terms=["A", "B", "C"]),
    )

    assert re.match(ptr, v.text(y, threshold=0.6))
    assert v.text(y, minimum_count=None, threshold=0.6) == ""

    assert v.text(x, join=",") == v.text(x).replace(";", ",")
    assert re.match(";".join([ptr] * 2), v.text(x, normalize=True))

    assert v.text([0] * 64) == "0.00F"
    assert v.text(v["D"].v) == "1.00D"


def test_capital(rng):
    v = Vocabulary(16, rng=rng)
    with pytest.raises(SpaParseError):
        v.parse("a")
    with pytest.raises(SpaParseError):
        v.parse("A+B+C+a")


def test_transform(rng):
    v1 = Vocabulary(32, rng=rng)
    v2 = Vocabulary(64, rng=rng)
    A = v1.parse("A")
    B = v1.parse("B")
    C = v1.parse("C")

    # Test transform from v1 to v2 (full vocbulary)
    # Expected: np.dot(t, A.v) ~= v2.parse('A')
    # Expected: np.dot(t, B.v) ~= v2.parse('B')
    # Expected: np.dot(t, C.v) ~= v2.parse('C')
    t = v1.transform_to(v2)

    assert v2.parse("A").compare(np.dot(t, A.v)) > 0.95
    assert v2.parse("C+B").compare(np.dot(t, C.v + B.v)) > 0.9

    # Test transform from v1 to v2 (only 'A' and 'B')
    t = v1.transform_to(v2, keys=["A", "B"])

    assert v2.parse("A").compare(np.dot(t, A.v)) > 0.95
    assert v2.parse("B").compare(np.dot(t, C.v + B.v)) > 0.95

    # Test transform_to when either vocabulary is read-only
    v1.parse("D")
    v2.parse("E")

    # When both are read-only, transform_to shouldn't add any new items to
    # either and the transform should be using keys that are the intersection
    # of both vocabularies
    v1.readonly = True
    v2.readonly = True

    t = v1.transform_to(v2)

    assert v1.keys == ["A", "B", "C", "D"]
    assert v2.keys == ["A", "B", "C", "E"]

    # When one is read-only, transform_to should add any new items to the non
    # read-only vocabulary
    v1.readonly = False
    v2.readonly = True

    t = v1.transform_to(v2)

    assert v1.keys == ["A", "B", "C", "D", "E"]
    assert v2.keys == ["A", "B", "C", "E"]

    # When one is read-only, transform_to should add any new items to the non
    # read-only vocabulary
    v1.readonly = True
    v2.readonly = False

    t = v1.transform_to(v2)

    assert v1.keys == ["A", "B", "C", "D", "E"]
    assert v2.keys == ["A", "B", "C", "E", "D"]


def test_prob_cleanup(rng):
    v = Vocabulary(64, rng=rng)
    assert 1.0 > v.prob_cleanup(0.7, 10000) > 0.9999
    assert 0.9999 > v.prob_cleanup(0.6, 10000) > 0.999
    assert 0.99 > v.prob_cleanup(0.5, 1000) > 0.9

    v = Vocabulary(128, rng=rng)
    assert 0.999 > v.prob_cleanup(0.4, 1000) > 0.997
    assert 0.99 > v.prob_cleanup(0.4, 10000) > 0.97
    assert 0.9 > v.prob_cleanup(0.4, 100000) > 0.8


def test_create_pointer_warning(rng):
    v = Vocabulary(2, rng=rng)

    # five pointers shouldn't fit
    with pytest.warns(UserWarning):
        v.parse("A")
        v.parse("B")
        v.parse("C")
        v.parse("D")
        v.parse("E")


def test_readonly(rng):
    v1 = Vocabulary(32, rng=rng)
    v1.parse("A+B+C")

    v1.readonly = True

    with pytest.raises(ValueError):
        v1.parse("D")


def test_subset(rng):
    v1 = Vocabulary(32, rng=rng)
    v1.parse("A+B+C+D+E+F+G")

    # Test creating a vocabulary subset
    v2 = v1.create_subset(["A", "C", "E"])
    assert v2.keys == ["A", "C", "E"]
    assert v2["A"] == v1["A"]
    assert v2["C"] == v1["C"]
    assert v2["E"] == v1["E"]
    assert v2.parent is v1

    # Test creating a subset from a subset (it should create off the parent)
    v3 = v2.create_subset(["C", "E"])
    assert v3.parent is v2.parent and v2.parent is v1

    v3.include_pairs = True
    assert v3.key_pairs == ["C*E"]
    assert not v1.include_pairs
    assert not v2.include_pairs

    # Test transform_to between subsets (should be identity transform)
    t = v1.transform_to(v2)

    assert v2.parse("A").compare(np.dot(t, v1.parse("A").v)) >= 0.99999999


def test_extend(rng, allclose):
    v = Vocabulary(16, rng=rng)
    v.parse("A+B")
    assert v.keys == ["A", "B"]
    assert not v.unitary

    # Test extending the vocabulary
    v.extend(["C", "D"])
    assert v.keys == ["A", "B", "C", "D"]

    # Test extending the vocabulary with various unitary options
    v.extend(["E", "F"], unitary=["E"])
    assert v.keys == ["A", "B", "C", "D", "E", "F"]
    assert v.unitary == ["E"]

    # Check if 'E' is unitary
    fft_val = np.fft.fft(v["E"].v)
    fft_imag = fft_val.imag
    fft_real = fft_val.real
    fft_norms = np.sqrt(fft_imag ** 2 + fft_real ** 2)
    assert allclose(fft_norms, np.ones(16))

    v.extend(["G", "H"], unitary=True)
    assert v.keys == ["A", "B", "C", "D", "E", "F", "G", "H"]
    assert v.unitary == ["E", "G", "H"]



import pytest

import nengo
from nengo.builder.probe import SimProbe
from nengo.exceptions import BuildError


def test_signal_probe(seed):
    """tests the function signal_probe"""

    # --- test probing wrong type
    with nengo.Network() as net:
        node = nengo.Node(0)
        p = nengo.Probe(node)

    with nengo.Simulator(net) as sim:
        probe_ops = [op for op in sim.model.operators if isinstance(op, SimProbe)]
        assert len(probe_ops) == 1
        assert probe_ops[0].signal is sim.model.sig[node]["out"]

    # hack to change probe target without tripping API validation
    nengo.Probe.target.data[p] = 1

    with pytest.raises(BuildError, match="Type .* is not probeable"):
        with nengo.Simulator(net):
            pass

    # --- test probing wrong attribute
    with nengo.Network() as net:
        p = nengo.Probe(nengo.Node(0))

    # hack to change probe attr without tripping API validation
    nengo.Probe.attr.data[p] = "batattr"

    with pytest.raises(BuildError, match="Attribute .* is not probeable"):
        with nengo.Simulator(net):
            pass



"""Configuration system to set defaults and backend-specific info.

The idea here is that a backend can create a Config and ConfigItems to
define the set of parameters that their backend supports.
Parameters are done as Python descriptors, so backends can also specify
error checking on those parameters.

A good writeup on descriptors (which has an example similar to Parameter)
can be found at
https://nbviewer.ipython.org/urls/gist.github.com/ChrisBeaumont/5758381/raw/descriptor_writeup.ipynb

"""

import inspect
import sys

from nengo.exceptions import ConfigError, ValidationError
from nengo.params import Default, is_param
from nengo.rc import rc
from nengo.utils.threading import ThreadLocalStack


class ClassParams:
    """A class to store extra parameters and defaults on configured classes.

    This is used by `.Config` to associate defaults and new `.Parameter`
    instances with existing objects. It should not be instantiated outside of
    `.Config.configures`.

    Parameters
    ----------
    configures : class
        The class with which to associate new defaults and parameters.
    """

    def __init__(self, configures):
        assert inspect.isclass(configures)
        self._configures = configures
        self._extra_params = {}
        self._default_params = tuple(
            attr
            for attr in dir(self._configures)
            if is_param(getattr(self._configures, attr))
        )

    def __contains__(self, key):
        return self in self.get_param(key)

    def __delattr__(self, key):
        if key.startswith("_"):
            super().__delattr__(key)
        else:
            self.get_param(key).del_default(self)

    def __getattr__(self, key):
        if key.startswith("_"):
            # If we get here, then that attribute hasn't been set
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{key}'"
            )
        return self.get_param(key).get_default(self)

    def __setattr__(self, key, value):
        """Overridden to handle instance descriptors manually.

        Everything not starting with _ is assumed to be a parameter.
        """
        if key.startswith("_"):
            super().__setattr__(key, value)
        else:
            param = self.get_param(key)
            if not param.configurable:
                raise ConfigError(f"Parameter '{key}' is not configurable")
            param.set_default(self, value)

    def __getstate__(self):
        state = {
            "_configures": self._configures,
            "_default_params": self._default_params,
            "_extra_params": self._extra_params,
        }

        # Store all of the things we set in the params
        for attr in self.params:
            param = self.get_param(attr)
            if self in param:
                state[attr] = param.get_default(self)

        return state

    def __setstate__(self, state):
        self._configures = state["_configures"]
        self._default_params = state["_default_params"]
        self._extra_params = state["_extra_params"]

        # Restore all of the things we set in the params
        for attr in self.params:
            if attr in state:
                self.get_param(attr).set_default(self, state[attr])

    def __str__(self):
        name = self._configures.__name__
        lines = [f"Parameters configured for {name}:"]
        for attr in self.params:
            if self in self.get_param(attr):
                lines.append(f"  {attr}: {getattr(self, attr)}")
        if len(lines) > 1:
            return "\n".join(lines)
        else:
            return f"No parameters configured for {name}."

    def __repr__(self):
        # Only print defaults if we've configured them
        params = []
        filled_defaults = [
            attr for attr in self.default_params if self in self.get_param(attr)
        ]
        for attr in filled_defaults + sorted(self.extra_params):
            params.append(f"{attr}: {getattr(self, attr)}")

        return (
            f"<{type(self).__name__}[{self._configures.__name__}]"
            f"{{{', '.join(params)}}}>"
        )

    @property
    def default_params(self):
        return self._default_params

    @property
    def extra_params(self):
        return tuple(self._extra_params)

    @property
    def params(self):
        return self.default_params + self.extra_params

    def get_param(self, key):
        if key in self._extra_params:
            return self._extra_params[key]

        return getattr(self._configures, key)

    def set_param(self, key, value):
        if not is_param(value):
            raise ConfigError(f"'{key}' is not a parameter")
        elif key in dir(self._configures):
            raise ConfigError(
                f"'{key}' is already a parameter in {self._configures.__name__}. "
                "Please choose a different name."
            )
        self._extra_params[key] = value

    def update(self, d):
        """Sets a number of parameters at once given a dictionary."""
        for key in d:
            setattr(self, key, d[key])


class InstanceParams:
    """A class to store parameter value on configured objects.

    In contrast to `.ClassParams`, the only thing that can be done with
    ``InstanceParams`` is get and set parameter values. Use the corresponding
    `.ClassParams` to set defaults and create new parameters.
    """

    def __init__(self, configures, clsparams):
        self._configures = configures
        self._clsparams = clsparams
        assert not inspect.isclass(configures)

    def __contains__(self, key):
        return self in self._clsparams.get_param(key)

    def __delattr__(self, key):
        if key.startswith("_"):
            super().__delattr__(key)
        elif key in dir(self._configures):
            # Disallow configuring attributes the instance already has
            raise ConfigError(
                f"Cannot configure the built-in parameter '{key}' on an instance of "
                f"'{type(self._configures).__name__}'. Please delete the attribute "
                "directly on the object."
            )
        else:
            self._clsparams.get_param(key).__delete__(self)

    def __getattr__(self, key):
        if key in self._clsparams.default_params:
            raise ConfigError(
                f"Cannot configure the built-in parameter '{key}' on an instance of "
                f"'{type(self._configures).__name__}'. Please get the attribute "
                "directly from the object."
            )
        param = self._clsparams.get_param(key)
        if self in param:
            return param.__get__(self, type(self))
        return getattr(self._clsparams, key)

    def __setattr__(self, key, value):
        """Everything not starting with _ is assumed to be a parameter."""
        if key.startswith("_"):
            super().__setattr__(key, value)
        elif key in dir(self._configures):
            # Disallow configuring attributes the instance already has
            raise ConfigError(
                f"Cannot configure the built-in parameter '{key}' on an instance of "
                f"'{type(self._configures).__name__}'. Please set the attribute "
                "directly on the object."
            )
        else:
            self._clsparams.get_param(key).__set__(self, value)

    def __getstate__(self):
        return dict(self.__dict__)

    def __setstate__(self, state):
        for k, v in state.items():
            setattr(self, k, v)

    def __repr__(self):
        params = []
        filled_params = [
            attr
            for attr in self._clsparams.params
            if self in self._clsparams.get_param(attr)
        ]
        for attr in filled_params:
            params.append(f"{attr}: {getattr(self, attr)}")

        return f"<{type(self).__name__}[{self._configures}]{{{', '.join(params)}}}>"

    def __str__(self):
        lines = [f"Parameters set for {str(self._configures)}:"]
        for attr in self._clsparams.params:
            if self in self._clsparams.get_param(attr):
                lines.append(f"  {attr}: {getattr(self, attr)}")
        return "\n".join(lines)

    def get_param(self, key):
        raise ConfigError(
            "Cannot get parameters on an instance; use "
            f"'config[{type(self._configures).__name__}].get_param' instead."
        )

    def set_param(self, key, value):
        raise ConfigError(
            "Cannot set parameters on an instance; use "
            f"'config[{type(self._configures).__name__}].set_param' instead."
        )


class Config:
    """Configures network-level defaults and additional parameters.

    Every `.Network` contains an associated ``Config`` object which can
    be manipulated to change network-specific defaults, and to store
    additional parameters (for example, those specific to a backend).

    A ``Config`` object can configure objects of any class, but it has to be
    told the classes to configure first. This is either done on instantiation
    of the ``Config`` object or by calling the `.configures`  method.
    This sets up a mapping between configured class and a `.ClassParams`
    object that sets the default values for that class. Attempting to
    configure an instance of a configure class will create a mapping from
    that instance to an `.InstanceParams` object to configure additional
    parameters for that instance.

    Parameters
    ----------
    *configures
        The classes that this ``Config`` instance will configure.

    Attributes
    ----------
    params : dict
        Maps configured classes and instances to their `.ClassParams`
        or `.InstanceParams` object.

    Examples
    --------
    To configure defaults on a network:

    .. testcode::

       net = nengo.Network()
       net.config[nengo.Ensemble].radius = 1.5
       with net:
           ens = nengo.Ensemble(10, 1)
       ens.radius == 1.5  # True

    To add a new parameter to a Nengo object:

    .. testcode::

       net.config[nengo.Ensemble].set_param(
           'location', nengo.params.Parameter('location')
       )
       net.config[ens].location = 'cortex'

    To group together a set of parameters:

    .. testcode::

       gaba = nengo.Config(nengo.Connection)
       gaba[nengo.Connection].synapse = nengo.Lowpass(0.008)
       with net, gaba:
           conn = nengo.Connection(ens, ens)
       conn.synapse == nengo.Lowpass(0.008)  # True

    To configure a new type of object:

    .. testcode::

       class SynapseInfo:
           label = nengo.params.StringParam('label', default=None)
       gaba.configures(SynapseInfo)
       gaba[SynapseInfo].label = "GABA"  # Set default label

    """

    context = ThreadLocalStack(maxsize=100)  # static stack of Config objects

    def __init__(self, *configures):
        self.params = {}
        if len(configures) > 0:
            self.configures(*configures)

    def __enter__(self):
        Config.context.append(self)
        return self

    def __exit__(self, dummy_exc_type, dummy_exc_value, dummy_tb):
        if len(Config.context) == 0:
            raise ConfigError(
                "Config.context in bad state; was empty when "
                "exiting from a 'with' block."
            )

        config = Config.context.pop()

        if config is not self:
            raise ConfigError(
                "Config.context in bad state; was expecting "
                f"current context to be '{self}' but instead got "
                f"'{config}'."
            )

    def __contains__(self, key):
        raise TypeError(f"Cannot check if {key!r} is in a config.")

    def __getitem__(self, key):
        # If we have the exact thing, we'll just return it
        if key in self.params:
            return self.params[key]

        # If key is a class return a superclass's ClassParams
        if inspect.isclass(key):
            for cls in key.__mro__:
                if cls in self.params:
                    return self.params[cls]

            # If no superclass ClassParams, KeyError
            raise ConfigError(
                f"Type '{key.__name__}' is not set up for configuration. "
                f"Call 'configures({key.__name__})' first."
            )

        # For new instances, if we configure a class in the mro we're good
        for cls in type(key).__mro__:
            if cls in self.params:
                clsparams = self.params[cls]
                instparams = InstanceParams(key, clsparams)
                self.params[key] = instparams
                return instparams

        # If we don't configure the class, KeyError
        raise ConfigError(
            f"Type '{type(key).__name__}' is not set up for configuration. Call "
            f"configures('{type(key).__name__}') first."
        )

    def __repr__(self):
        classes = [key.__name__ for key in self.params if inspect.isclass(key)]
        return (
            f"<{type(self).__name__} at 0x{id(self):x} "
            f"configuring ({', '.join(classes)})>"
        )

    def __str__(self):
        return "\n".join(str(v) for v in self.params.values())

    @staticmethod
    def all_defaults(nengo_cls=None):
        """Look up all of the default values in the current context.

        Parameters
        ----------
        nengo_cls : class, optional
            If specified, only the defaults for a particular class will
            be returned. If not specified, the defaults for all configured
            classes will be returned.

        Returns
        -------
        str
        """
        lines = []
        if nengo_cls is None:
            all_configured = set()
            for config in Config.context:
                all_configured.update(
                    key for key in config.params if inspect.isclass(key)
                )
            lines.extend([Config.all_defaults(key) for key in all_configured])
        else:
            lines.append(f"Current defaults for {nengo_cls.__name__}:")
            for attr in dir(nengo_cls):
                desc = getattr(nengo_cls, attr)
                if is_param(desc) and desc.configurable:
                    val = Config.default(nengo_cls, attr)
                    lines.append(f"  {attr}: {val}")
        return "\n".join(lines)

    @staticmethod
    def default(nengo_cls, param):
        """Look up the current default value for a parameter.

        The default is found by going through the config stack, from most
        specific to least specific. The network that an object is in
        is the most specific; the top-level network is the least specific.
        If no default is found there, then the parameter's default value
        is returned.
        """

        # Get the descriptor
        desc = getattr(nengo_cls, param)
        if not desc.configurable:
            raise ConfigError(
                "Unconfigurable parameters have no defaults. "
                "Please ensure you are not using the 'Default' "
                "keyword with an unconfigurable parameter."
            )

        for config in reversed(Config.context):
            for cls in nengo_cls.__mro__:
                # If a default has been set for this config, return it
                if cls in config.params and config[cls] in desc:
                    return getattr(config[cls], param)

        # Otherwise, return the param default
        return desc.default

    def configures(self, *classes):
        """Start configuring a particular class and its instances."""
        if len(classes) == 0:
            raise TypeError("configures() takes 1 or more arguments (0 given)")
        for klass in classes:
            self.params[klass] = ClassParams(klass)


class SupportDefaultsMixin:
    """Mixin to support assigning ``Default`` to parameters.

    Implements ``__setattr__`` to do so. If the inheriting class overrides
    this method, it has to call the mixin's ``__setattr__``.

    This mixin may simplify the exception depending on the value of the
    ``simplified`` rc option.
    """

    def __setattr__(self, name, val):
        if val is Default:
            val = Config.default(type(self), name)

        if rc["exceptions"].getboolean("simplified"):
            try:
                super().__setattr__(name, val)
            except ValidationError as e:
                exc_info = sys.exc_info()
                raise exc_info[1].with_traceback(None) from e
        else:
            super().__setattr__(name, val)
